def solve_one_root_quadratic(a,b,c):
    """
    Obtain one root of the quadratic equation of the form ax^2 + bx + c = 0.

    Parameters
    ----------
    a : float
        Coefficient of x^2
    b : float
        Coefficient of x
    c : float
        Constant term

    Returns
    -------
    float
        One root of the quadratic equation
    """
    return 2*c / (-b - np.sqrt(b**2 - 4*a*c))


def solve_one_root_depressed_cubic(p,q):

    """
    Obtain one root of the depressed cubic equation of the form x^3 + p x + q = 0.

    Parameters
    ----------
    p : float
        Coefficient of x
    q : float
        Constant term

    Returns
    -------
    float
        One real root of the cubic equation
    """

    delta = np.sqrt((q**2/4) + (p**3/27))

    return np.cbrt(-q/2+delta) + np.cbrt(-q/2-delta)

def residuals_squares_sum(y_true,y_pred):

    """
    Calculate the residual sum of squares.

    Parameters
    ----------
    y_true : array-like
        True values
    y_pred : array-like
        Predicted values

    Returns
    -------
    float
        Residual sum of squares
    """

    # Convert to numpy arrays if it is a list
    if isinstance(y_true, list):
        y_true = np.array(y_true)

    if isinstance(y_pred, list):
        y_pred = np.array(y_pred)

    rss = np.sum((y_true - y_pred)**2)

    return rss



def r_squared(y_true, y_pred):
    """
    Calculate the R-squared value for a regression model.

    Parameters
    ----------
    y_true : array-like
        True values
    y_pred : array-like
        Predicted values

    Returns
    -------
    float
        R-squared value
    """

    ss_total = np.sum((y_true - np.mean(y_true))**2)
    ss_res = np.sum((y_true - y_pred)**2)
    return 1 - ss_res / ss_total


def adjusted_r2(r2, n, p):
    """
    Calculate the adjusted R-squared value for a regression model.

    Parameters
    ----------
    r2 : float
        R-squared value
    n : int
        Number of observations
    p : int
        Number of predictors

    Returns
    -------
    float
        Adjusted R-squared value
    """

    return 1 - (1 - r2) * (n - 1) / (n - p - 1)

def compute_aic(y_true, y_pred, k):
    """
    Compute the Akaike Information Criterion (AIC) for a regression model.

    Parameters
    ----------
    y_true : array-like
        True values
    y_pred : array-like
        Predicted values
    k : int
        Number of parameters in the model

    Returns
    -------
    float
        AIC value
    """

    n = len(y_true)
    rss = np.sum((y_true - y_pred) ** 2)
    return n * np.log(rss / n) + 2 * k


def compare_akaikes(akaikes_1, akaikes_2, akaikes_3, akaikes_4, denaturant_concentrations):
    model_names = ['Linear - Linear', 'Linear - Quadratic',
                   'Quadratic - Linear', 'Quadratic - Quadratic']

    akaikes_df = pd.DataFrame({
        'Model': model_names})

    i = 0
    for a1, a2, a3, a4 in zip(akaikes_1, akaikes_2, akaikes_3, akaikes_4):
        # Create a new column with the Akaike values
        # The name is the denaturant concentration

        # Compute delta AIC
        min_aic = np.min([a1, a2, a3, a4])
        a1 = a1 - min_aic
        a2 = a2 - min_aic
        a3 = a3 - min_aic
        a4 = a4 - min_aic

        akaikes_df[str(i) + '_' + str(denaturant_concentrations[i])] = [a1, a2, a3, a4]
        i += 1

    # Find the best model for each denaturant concentration
    best_models_ids = []
    for i in range(len(denaturant_concentrations)):

        # Get the column with the Akaike values
        aic_col = akaikes_df.iloc[:, i + 1].to_numpy()

        # Find index that sort them from min to max a numpy array
        sorted_idx = np.argsort(aic_col)

        first_model_id = np.arange(4)[sorted_idx][0]
        second_model_id = np.arange(4)[sorted_idx][1]
        third_model_id = np.arange(4)[sorted_idx][2]
        fourth_model_id = np.arange(4)[sorted_idx][3]

        best_models_ids.append(first_model_id)

        # Compare the AIC value of the second model to the first one
        if aic_col[second_model_id] - aic_col[first_model_id] < 2:
            best_models_ids.append(second_model_id)

        # Compare the AIC value of the third model to the first one
        if aic_col[third_model_id] - aic_col[first_model_id] < 2:
            best_models_ids.append(third_model_id)

        # Compare the AIC value of the fourth model to the first one
        if aic_col[fourth_model_id] - aic_col[first_model_id] < 2:
            best_models_ids.append(fourth_model_id)

    # Print the overall best model
    best_model_all = Counter(best_models_ids).most_common(1)[0][0]
    return model_names[best_model_all]


def rss_p(rrs0, n, p, alfa):

    """
    Given the residuals of the best fitted model,
    compute the desired residual sum of squares for a 1-alpha confidence interval.
    This is used to compute asymmetric confidence intervals for the fitted parameters.

    Parameters
    ----------
    rrs0 : float
        Residual sum of squares of the model with the best fit
    n : int
        Number of data points
    p : int
        Number of parameters
    alfa : float
        Desired significance level (alpha)

    Returns
    -------
    float
        Residual sum of squares for the desired confidence interval
    """

    critical_value = stats.f.ppf(q=1 - alfa, dfn=1, dfd=n - p)

    return rrs0 * (1 + critical_value / (n - p))


def get_desired_rss(y, y_fit, p,alpha=0.05):

    """
    Given the observed and fitted data, find the residual sum of squares required for a 1-alpha confidence interval.

    Parameters
    ----------
    y : array-like
        Observed values or list of arrays
    y_fit : array-like
        Fitted values or list of arrays
    p : int
        Number of parameters
    alpha : float, optional
        Desired significance level (default: 0.05)

    Returns
    -------
    float
        Residual sum of squares corresponding to the desired confidence interval
    """

    # If y is of type list, convert it to a numpy array by concatenating
    if isinstance(y, list):
        y = np.concatenate(y,axis=0)
    # If y_fit is of type list, convert it to a numpy array by concatenating
    if isinstance(y_fit, list):
        y_fit = np.concatenate(y_fit,axis=0)

    n = len(y)

    rss = get_rss(y, y_fit)

    return rss_p(rss, n, p, alpha)

def compare_linear_to_quadratic(x,y):

    """
    Compare the linear and quadratic fits to the data using an F-test.

    Parameters
    ----------
    x : array-like
        x data
    y : array-like
        y data

    Returns
    -------
    bool
        True if the linear model is statistically preferable to the quadratic model
    """

    m, b       = fit_line_robust(x, y)
    y_pred_lin = m * x + b

    a,b,c     = fit_quadratic_robust(x, y)
    y_pred_quad = a * x ** 2 + b * x + c

    # Residual sums
    rss_lin = np.sum((y - y_pred_lin) ** 2)
    rss_quad = np.sum((y - y_pred_quad) ** 2)

    # R² and Adjusted R²
    n = len(x)
    p_lin = 1
    p_quad = 2

    # F-test
    numerator   = (rss_lin - rss_quad) / (p_quad - p_lin)
    denominator = rss_quad / (n - (p_quad + 1))
    f_stat = numerator / denominator
    p_value = 1 - f_dist.cdf(f_stat, dfn=p_quad - p_lin, dfd=n - (p_quad + 1))

    # True if linear model is better
    return p_value > 0.05

def fu_two_state_dimer(K,C):
    """
    Given the equilibrium constant K of N2 <-> 2U and the concentration of dimer equivalent C,
    return the fraction of unfolded protein.

    Parameters
    ----------
    K : float
        Equilibrium constant of the reaction N2 <-> 2U
    C : float
        Concentration of dimer equivalent

    Returns
    -------
    float
        Fraction of unfolded protein
    """

    return solve_one_root_quadratic(4*C, K, -K)

def arrhenius(T, Tf, Ea):
    """
    Arrhenius equation: defines dependence of reaction rate constant k on temperature.
    In this version of the equation we use Tf (a temperature of k=1) to avoid specifying a pre-exponential constant A.

    Parameters
    ----------
    T : array-like
        Temperature (°C or K)
    Tf : float
        Reference temperature at which the reaction rate constant equals 1 (°C or K)
    Ea : float
        Activation energy (kcal/mol)

    Returns
    -------
    numpy.ndarray
        Reaction rate constant at the given temperature
    """

    T  = temperature_to_kelvin(T)
    Tf = temperature_to_kelvin(Tf)

    return np.exp(-Ea / R_gas * (1 / T - 1 / Tf))


def fit_tc_unfolding_many_signals_slow(
        list_of_temperatures,
        list_of_signals,
        signal_ids,
        denaturant_concentrations,
        initial_parameters,
        low_bounds, high_bounds,
        signal_fx,
        fit_slope_native_temp=True,
        fit_slope_unfolded_temp=True,
        fit_slope_native_den=True,
        fit_slope_unfolded_den=True,
        fit_quadratic_native=False,
        fit_quadratic_unfolded=False,
        oligomer_concentrations=None,
        fit_m1=False,
        model_scale_factor=False,
        scale_factor_exclude_ids=[]):
    """
    Fit thermochemical unfolding curves for many signals (slow variant).

    Parameters
    ----------
    list_of_temperatures : list of array-like
        List of temperature arrays for each dataset
    list_of_signals : list of array-like
        List of signal arrays for each dataset
    signal_ids : list of int
        Signal-type id for each dataset (0..n_signals-1)
    denaturant_concentrations : list
        Denaturant concentrations for each dataset (flattened across signals)
    initial_parameters : array-like
        Initial guess for the parameters
    low_bounds : array-like
        Lower bounds for the parameters
    high_bounds : array-like
        Upper bounds for the parameters
    signal_fx : callable
        Signal model function
    fit_slope_native_temp : bool, optional
        Whether to fit the temperature slope of the native baseline (per-signal)
    fit_slope_unfolded_temp : bool, optional
        Whether to fit the temperature slope of the unfolded baseline (per-signal)
    fit_slope_native_den : bool, optional
        Whether to fit the denaturant slope of the native baseline (per-signal)
    fit_slope_unfolded_den : bool, optional
        Whether to fit the denaturant slope of the unfolded baseline (per-signal)
    fit_quadratic_native : bool, optional
        Whether to fit a quadratic temperature term for the native baseline (per-signal)
    fit_quadratic_unfolded : bool, optional
        Whether to fit a quadratic temperature term for the unfolded baseline (per-signal)
    oligomer_concentrations : list, optional
        Oligomer concentrations per dataset (used by oligomeric models)
    fit_m1 : bool, optional
        Whether to include and fit temperature dependence of the m-value (m1)
    model_scale_factor : bool, optional
        If True, include a per-denaturant concentration scale factor to account for intensity differences
    scale_factor_exclude_ids : list, optional
        IDs of scale factors to exclude / fix to 1 (useful to avoid fitting trivial factors)

    Returns
    -------
    global_fit_params : numpy.ndarray
        Fitted global parameters
    cov : numpy.ndarray
        Covariance matrix of the fitted parameters
    predicted_lst : list of numpy.ndarray
        Predicted signals for each dataset based on the fitted parameters
    """

    all_signal = np.concatenate(list_of_signals, axis=0)

    n_signals = np.max(signal_ids) + 1

    nr_den = int(len(denaturant_concentrations) / n_signals)

    if len(scale_factor_exclude_ids) > 0 and model_scale_factor:
        # Sort them in ascending order to avoid issues when inserting
        scale_factor_exclude_ids = sorted(scale_factor_exclude_ids)

    # Find if highest concentration of denaturant has a higher signal or not
    if model_scale_factor:
        den_conc_simple = denaturant_concentrations[:nr_den]

        # Find the index that sorts them in descending order from highest to lowest
        sort_indeces = np.argsort(den_conc_simple)[::-1]

        signal_first = list_of_signals[:nr_den]

        signal_sort = [signal_first[i] for i in sort_indeces]

        higher_den_equal_higher_signal = signal_sort[0][0] > signal_sort[-1][0]

    def unfolding(dummyVariable, *args):

        Tm, DHm, Cp0, m0 = args[:4]  # Enthalpy of unfolding, Temperature of melting, Cp0, m0, m1

        id_param_init = 4 + fit_m1
        m1 = args[4] if fit_m1 else 0

        # First filter, verify that DG is not lower than 0 at 5C
        # In other words, we do not have cold denaturation at 5C
        """
        Tfive = temperature_to_kelvin(5)
        TmK   = temperature_to_kelvin(Tm)

        DGfive = DHm * (1 - Tfive / TmK) + Cp0 * (Tfive - TmK - Tfive * np.log(Tfive / TmK))

        if DGfive < 0:

            return np.zeros(len(all_signal))
        """

        a_Ns = args[id_param_init:id_param_init + n_signals]
        a_Us = args[id_param_init + n_signals:id_param_init + 2 * n_signals]

        id_param_init = id_param_init + 2 * n_signals
        if fit_slope_native_temp:
            b_Ns = args[id_param_init:id_param_init + n_signals]
            id_param_init += n_signals
        else:
            b_Ns = [0] * n_signals

        if fit_slope_unfolded_temp:
            b_Us = args[id_param_init:id_param_init + n_signals]
            id_param_init += n_signals
        else:
            b_Us = [0] * n_signals

        if fit_slope_native_den:
            c_Ns = args[id_param_init:id_param_init + n_signals]
            id_param_init += n_signals
        else:
            c_Ns = [0] * n_signals

        if fit_slope_unfolded_den:
            c_Us = args[id_param_init:id_param_init + n_signals]
            id_param_init += n_signals
        else:
            c_Us = [0] * n_signals

        if fit_quadratic_native:
            d_Ns = args[id_param_init:id_param_init + n_signals]
            id_param_init += n_signals
        else:
            d_Ns = [0] * n_signals

        if fit_quadratic_unfolded:
            d_Us = args[id_param_init:id_param_init + n_signals]
            id_param_init += n_signals
        else:
            d_Us = [0] * n_signals

        if model_scale_factor:
            # One per denaturant concentration
            factors = args[id_param_init:id_param_init + (nr_den - len(scale_factor_exclude_ids))]

            for id_ex in scale_factor_exclude_ids:
                factors = np.insert(factors, id_ex, 1)

            # Repeat the list so have the same length as list_of_temperatures, equal to denaturant concentration * number of signals
            factors = np.tile(factors, n_signals)

            id_param_init += nr_den

        signal = []

        for i, T in enumerate(list_of_temperatures):
            a_N = a_Ns[signal_ids[i]]
            b_N = b_Ns[signal_ids[i]]
            c_N = c_Ns[signal_ids[i]]
            d_N = d_Ns[signal_ids[i]]

            a_U = a_Us[signal_ids[i]]
            b_U = b_Us[signal_ids[i]]
            c_U = c_Us[signal_ids[i]]
            d_U = d_Us[signal_ids[i]]

            d = denaturant_concentrations[i]

            c = 0 if oligomer_concentrations is None else oligomer_concentrations[i]

            d_factor = 1

            d = d * d_factor

            y = signal_fx(
                T, d, DHm, Tm, Cp0, m0, m1,
                a_N, b_N, c_N, d_N,
                a_U, b_U, c_U, d_U, c
            )

            scale_factor = 1 if not model_scale_factor else factors[i]

            y = y * scale_factor

            signal.append(y)

        # Second filter, verify that higher_den_equal_higher_signal is same in the raw and fitted signal
        if model_scale_factor:

            signal_first = signal[:nr_den]

            signal_sort = [signal_first[i] for i in sort_indeces]

            pred_higher_den_equal_higher_signal = signal_sort[0][0] > signal_sort[-1][0]

            if pred_higher_den_equal_higher_signal != higher_den_equal_higher_signal:
                return np.zeros(len(all_signal))

        return np.concatenate(signal, axis=0)

    global_fit_params, cov = curve_fit(
        unfolding, 1, all_signal,
        p0=initial_parameters,
        bounds=(low_bounds, high_bounds))

    predicted = unfolding(1, *global_fit_params)

    # Convert predict to list of lists
    predicted_lst = []

    init = 0
    for T in list_of_temperatures:
        n = len(T)
        predicted_lst.append(predicted[init:init + n])
        init += n

    return global_fit_params, cov, predicted_lst


import numpy as np
import pytest

from pychemelt.utils.processing import (
    guess_Tm_from_derivative,
    get_colors_from_numeric_values,
    fit_local_thermal_unfolding_to_signal_lst,
    adjust_value_to_interval,
    parse_number,
    are_all_strings_numeric,
    is_float,
    transform_to_list
)


from pychemelt.utils.palette import VIRIDIS

def test_error_not_enough_data():

    temp_lst = [[1,2,3]]
    deriv_lst = [[1,2,3]]

    x1 = 0
    x2 = 5

    with pytest.raises(ValueError):

        guess_Tm_from_derivative(temp_lst, deriv_lst, x1, x2)

def test_get_colors_from_numeric_values():

    y = [1, 2, 3, 4, 5]

    colors = get_colors_from_numeric_values(y, 1, 5,use_log_scale=True)

    assert colors[0] == VIRIDIS[0]
    assert colors[-1] == VIRIDIS[-1]

    colors = get_colors_from_numeric_values(y, 1, 5,use_log_scale=False)

    assert colors[0] == VIRIDIS[0]
    assert colors[-1] == VIRIDIS[-1]

def test_trigger_exception_fit_local_thermal_unfolding_to_signal_lst():

    signal_lst = [[np.nan for _ in range(5)]]
    temp_lst = [[x for x in range(5)]]

    Tms, dHs, predicted_lst =  fit_local_thermal_unfolding_to_signal_lst(
        signal_lst, temp_lst, [100],
        [1], [1], [1], [1], [1], [1],
        baseline_native_fx = lambda x: x,
        baseline_unfolded_fx = lambda x: x
    )

    assert Tms == []

def test_adjust_value_to_interval():

    assert adjust_value_to_interval(10, 0, 100,1) == 10

    assert adjust_value_to_interval(-1, 0, 100,0.5) == 0.5

    assert adjust_value_to_interval(110, 0, 100,0.5) == 99.5

def test_parse_number():

    assert parse_number('10') == 10

    assert parse_number('10.5') == 10.5

    assert parse_number('1e-3') == 0.001

    with pytest.raises(ValueError):
        parse_number('not_a_number')

def test_are_all_strings_numeric():

    assert are_all_strings_numeric(['10', '20', '30']) == True

    assert are_all_strings_numeric(['10', '20.5', '1e-3']) == True

    assert are_all_strings_numeric(['10', 'not_a_number', '30']) == False

def test_is_float():

    assert is_float('10') == True

    assert is_float('10.5') == True

    assert is_float('1e-3') == True

    assert is_float('not_a_number') == False

def transform_to_list():

    assert transform_to_list("hola") == ["hola"]

    assert transform_to_list([1,2,3]) == [1,2,3]

    assert transform_to_list('string') == ['string']

    # none should be returned as none, not as a list with one element
    assert transform_to_list(None) == None

    # raise error if input is not string, list, float, int or None
    with pytest.raises(ValueError):
        transform_to_list({'key': 'value'})




"""
This module contains helper functions for mathematical operations
Author: Osvaldo Burastero
"""


import numpy as np

from .constants    import Tref_cst
from scipy.signal  import savgol_filter

__all__ = [
    "temperature_to_kelvin",
    "temperature_to_celsius",
    "shift_temperature",
    "constant_baseline",
    "linear_baseline",
    "quadratic_baseline",
    "exponential_baseline",
    "is_evenly_spaced",
    "first_derivative_savgol",
    "relative_errors",
    "find_line_outliers",
    "get_rss",
    "solve_one_root_quadratic",
    "solve_one_root_depressed_cubic",
]

def temperature_to_kelvin(T):
    """
    Convert temperature from Celsius to Kelvin if necessary.

    Parameters
    ----------
    T : array-like
        Temperature values

    Returns
    -------
    array-like
        Temperature values in Kelvin
    """
    return T + 273.15 if np.max(T) < 270 else T

def temperature_to_celsius(T):
    """
    Convert temperature from Kelvin to Celsius if necessary.

    Parameters
    ----------
    T : array-like
        Temperature values

    Returns
    -------
    array-like
        Temperature values in Celsius
    """
    return T - 273.15 if np.max(T) > 270 else T

def shift_temperature(T):
    """
    Shift temperature to be relative to Tref_cst in Kelvin.

    Parameters
    ----------
    T : array-like
        Temperature values

    Returns
    -------
    array-like
        Shifted temperature values
    """
    return temperature_to_kelvin(T) - Tref_cst

def shift_temperature_K(T):
    """
    Shift temperature in Kelvin to be relative to Tref_cst.

    Parameters
    ----------
    T : array-like
        Temperature values in Kelvin

    Returns
    -------
    array-like
        Shifted temperature values
    """
    return T - Tref_cst

def constant_baseline(dt,d,den_slope,a,*args):

    """
    Baseline function with no dependence on temperature and dependence on denaturant concentration

    Parameters
    ----------
    dt : float
        delta temperature, not used here but required for compatibility with other baseline functions
    d : float
        denaturant concentration
    den_slope : float
        linear dependence of signal on denaturant concentration
    a : float
        intercept of the baseline

    Returns
    ------
    float
        Baseline signal
    """

    return a + den_slope * d

def linear_baseline(dt,d,den_slope,a,b,*args):

    """
    Baseline function with linear dependence on temperature and linear dependence on denaturant concentration

    Parameters
    ----------
    dt : float
        delta temperature, not used here but required for compatibility with other baseline functions
    d : float
        denaturant concentration
    den_slope : float
        linear dependence of signal on denaturant concentration
    a : float
        intercept of the baseline
    b : float
        linear dependence of signal on temperature

    Returns
    ------
    float
        Baseline signal
    """

    return a + b*dt + den_slope * d

def quadratic_baseline(dt,d,den_slope,a,b,c):

    """
    Baseline function with quadratic dependence on temperature and linear dependence on denaturant concentration

    Parameters
    ----------
    dt : float
        delta temperature, not used here but required for compatibility with other baseline functions
    d : float
        denaturant concentration
    den_slope : float
        linear dependence of signal on denaturant concentration
    a : float
        intercept of the baseline
    b : float
        linear dependence of signal on temperature
    c : float
        quadratic dependence of signal on temperature

    Returns
    ------
    float
        Baseline signal
    """

    return a + b*dt + c*dt**2 + den_slope * d

def exponential_baseline(dt,d,den_slope,a,c,alpha):

    """
    Baseline function with exponential dependence on temperature and linear dependence on denaturant concentration

    Parameters
    ----------
    dt : float
        delta temperature, not used here but required for compatibility with other baseline functions
    d : float
        denaturant concentration
    den_slope : float
        linear dependence of signal on denaturant concentration
    a : float
        intercept of the baseline
    b : float
        pre-exponential factor for the dependence on temperature
    c : float
        exponential coefficient for the dependence on temperature

    Returns
    ------
    float
        Baseline signal
    """

    return a + c * np.exp(-alpha * dt) + den_slope * d

def is_evenly_spaced(x, tol = 1e-4):
    """
    Check if x is evenly spaced within a given tolerance.

    Parameters
    ----------
    x : array-like
        x data
    tol : float, optional
        Tolerance for considering spacing equal (default: 1e-4)

    Returns
    -------
    bool
        True if x is evenly spaced, False otherwise
    """

    diffs = np.diff(x)

    return np.all(np.abs(diffs - diffs[0]) < tol)


def first_derivative_savgol(x, y, window_length=5, polyorder=4):

    """
    Estimate the first derivative using Savitzky-Golay filtering.

    Parameters
    ----------
    x : array-like
        x data (must be evenly spaced)
    y : array-like
        y data
    window_length : int, optional
        Length of the filter window, in temperature units (default: 5)
    polyorder : int, optional
        Order of the polynomial used to fit the samples (default: 4)

    Returns
    -------
    numpy.ndarray
        First derivative of y with respect to x

    Notes
    -----
    This function will raise a ValueError if `x` is not evenly spaced.
    """

    # Check if x is evenly spaced
    if not is_evenly_spaced(x):
        raise ValueError("x must be evenly spaced for Savitzky-Golay filter.")

    # Calculate spacing (assuming uniform x)
    dx = np.mean(np.diff(x))
    odd_n_data_points_window_len = np.ceil(window_length / dx) // 2 * 2 + 1

    if polyorder >= odd_n_data_points_window_len:
        # Raise error we need more data points for polynomial fit
        raise ValueError("polyorder must be less than window_length.")

    # Apply Savitzky-Golay filter for first derivative
    dydx = savgol_filter(y, window_length=int(odd_n_data_points_window_len), polyorder=polyorder, deriv=1, delta=dx, mode="nearest")

    return dydx


def relative_errors(params,cov):
    """
    Calculate the relative errors of the fitted parameters.

    Parameters
    ----------
    params : numpy.ndarray
        Fitted parameters
    cov : numpy.ndarray
        Covariance matrix of the fitted parameters

    Returns
    -------
    numpy.ndarray
        Relative errors of the fitted parameters (in percent)
    """

    error = np.sqrt(np.diag(cov))
    rel_error = np.abs(error / params) * 100

    return rel_error

def find_line_outliers(m,b,x,y,sigma=2.5):
    """
    Find outliers in a linear fit using the sigma rule.

    Parameters
    ----------
    m : float
        Slope of the line
    b : float
        Intercept of the line
    x : array-like
        x data
    y : array-like
        y data
    sigma : float, optional
        Number of standard deviations to use for outlier detection (default: 2.5)

    Returns
    -------
    numpy.ndarray
        Indices of the outliers
    """

    # Calculate the residuals
    residuals = y - (m * x + b)

    # Calculate the standard deviation of the residuals
    std_residuals = np.std(residuals)

    # Calculate the mean of the residuals
    mean_residuals = np.mean(residuals)

    # Identify outliers
    outliers = np.where(np.abs(residuals - mean_residuals) > sigma * std_residuals)[0]

    return outliers

def get_rss(y, y_fit):

    """
    Compute the residual sum of squares.

    Parameters
    ----------
    y : array-like
        Observed values
    y_fit : array-like
        Fitted values

    Returns
    -------
    float
        Residual sum of squares
    """

    residuals = y - y_fit
    rss       = np.sum(residuals ** 2)

    return rss


def solve_one_root_quadratic(a,b,c):
    """
    Solution to one root quadratic: a * X**2 + b * X + c = 0
    
    Parameters
    ----------
    a : number type
        parameter a
    b : number type
        parameter b
    c : number type
        parameter c

    Returns
    -------
    float
        Solution of the formula
    """
    return 2*c / (-b - np.sqrt(b**2 - 4*a*c))

def solve_one_root_depressed_cubic(p,q):
    """
    Solution to one root depressed cubic: X**3 + p * X + q = 0
    
    Parameters
    ----------
    p : number type
        parameter p
    q : number type
        parameter q

    Returns
    -------
    float
        Solution of the formula
    """

    delta = np.sqrt((q**2/4) + (p**3/27))

    return np.cbrt(-q/2+delta) + np.cbrt(-q/2-delta)


import numpy as np

from pychemelt.utils.fitting import (
    fit_thermal_unfolding,
    fit_tc_unfolding_single_slopes,
    fit_tc_unfolding_shared_slopes_many_signals,
    fit_tc_unfolding_many_signals
)

from pychemelt.utils.math import exponential_baseline

from pychemelt.utils.signals import (
    signal_two_state_t_unfolding,
    signal_two_state_tc_unfolding
)

# Centralized test constants
RNG_SEED = 2
TEMP_START = 30.0
TEMP_STOP = 90.0
N_TEMPS = 80
CONCS = [0.01, 1, 2, 2.6, 3, 4, 5]

# Model / ground-truth parameters
DHm_VAL = 120
Tm_VAL = 65
CP0_VAL = 1.8
M0_VAL = 2.6
M1_VAL = 0

INTERCEPT_N = 100
PRE_EXP_N = 1
C_N_VAL = 0
ALPHA_N_VAL = 0.1
INTERCEPT_U = 110
PRE_EXP_U = 1
C_U_VAL = 0
ALPHA_U_VAL = 0.2

rng = np.random.default_rng(RNG_SEED)

def_params = {
    'DHm': DHm_VAL,
    'Tm': Tm_VAL+273.15,
    'Cp0': CP0_VAL,
    'm0': M0_VAL,
    'm1': M1_VAL,
    'p1_N': C_N_VAL,
    'p2_N': INTERCEPT_N,
    'p3_N': PRE_EXP_N,
    'p4_N': ALPHA_N_VAL,
    'p1_U': C_U_VAL,
    'p2_U': INTERCEPT_U,
    'p3_U': PRE_EXP_U,
    'p4_U': ALPHA_U_VAL,
    'baseline_N_fx':exponential_baseline,
    'baseline_U_fx':exponential_baseline

}

concs = CONCS

# Calculate signal range for proper y-axis scaling
temp_range  = np.linspace(TEMP_START, TEMP_STOP, N_TEMPS)
temp_range_K = temp_range + 273.15
signal_list = []
temp_list   = []

for D in concs:

    y = signal_two_state_tc_unfolding(temp_range_K, D, **def_params)

    # Add gaussian error to signal
    y += rng.normal(0, 0.005, len(y))

    signal_list.append(y)
    temp_list.append(temp_range)


def test_fit_thermal_unfolding_exponential():

    p0 = [Tm_VAL - 5, INTERCEPT_N] + [1]*6
    low_bounds = [TEMP_START, TEMP_START]   + [-np.inf]*6
    high_bounds = [TEMP_STOP, 200] + [np.inf]*6

    global_fit_params, cov, predicted_lst = fit_thermal_unfolding(
        list_of_temperatures=temp_list[:1],
        list_of_signals=signal_list[:1],
        initial_parameters=p0,
        low_bounds=low_bounds,
        high_bounds=high_bounds,
        signal_fx=signal_two_state_t_unfolding,
        baseline_native_fx=exponential_baseline,
        baseline_unfolded_fx=exponential_baseline,
        Cp=0,
    )

    expected = [Tm_VAL, DHm_VAL]

    np.testing.assert_allclose(global_fit_params[:2], expected, rtol=0.1, atol=0)


def test_fit_tc_unfolding_single_slopes_exponential():

    p0 = [Tm_VAL, DHm_VAL, CP0_VAL, M0_VAL] + [0.1]*(6*len(concs))
    low_bounds = [TEMP_START, TEMP_START, 1, 1]   + [1e-5]*(6*len(concs))
    high_bounds = [TEMP_STOP, 200, 5, 5] + [1e3]*(6*len(concs))

    kwargs = {
        'list_of_temperatures' : temp_list,
        'list_of_signals' : signal_list,
        'denaturant_concentrations' : concs,
        'signal_fx' : signal_two_state_tc_unfolding,
        'baseline_native_fx':exponential_baseline,
        'baseline_unfolded_fx':exponential_baseline
    }

    global_fit_params, cov, predicted_lst = fit_tc_unfolding_single_slopes(
        initial_parameters=p0,
        low_bounds=low_bounds,
        high_bounds=high_bounds,
        **kwargs
    )

    expected = [Tm_VAL, DHm_VAL, CP0_VAL, M0_VAL]

    np.testing.assert_allclose(global_fit_params[:4], expected, rtol=0.1, atol=0)

    # Fit with fixed Tm
    p0_tm = p0.copy()
    low_bounds_tm = low_bounds.copy()
    high_bounds_tm = high_bounds.copy()

    p0_tm.pop(0)
    low_bounds_tm.pop(0)
    high_bounds_tm.pop(0)

    global_fit_params, cov, predicted_lst = fit_tc_unfolding_single_slopes(
        initial_parameters=p0_tm,
        low_bounds=low_bounds_tm,
        high_bounds=high_bounds_tm,
        tm_value=Tm_VAL,
        **kwargs
    )

    expected = [DHm_VAL, CP0_VAL, M0_VAL]

    np.testing.assert_allclose(global_fit_params[:3], expected, rtol=0.1, atol=0)

    # End of - Fit with fixed Tm

    # Fit with fixed DH and fixed Tm and fixed Cp - allow fit of m1
    p0_2 = p0_tm[2:]
    low_bounds_2 = low_bounds_tm[2:]
    high_bounds_2 = high_bounds_tm[2:]

    p0_2.insert(1, 0)
    low_bounds_2.insert(1, -0.5)
    high_bounds_2.insert(1, 0.5)

    global_fit_params, cov, predicted_lst = fit_tc_unfolding_single_slopes(
        initial_parameters=p0_2,
        low_bounds=low_bounds_2,
        high_bounds=high_bounds_2,
        dh_value=DHm_VAL,
        tm_value=Tm_VAL,
        cp_value=CP0_VAL,
        fit_m1=True,
        **kwargs
    )

    np.testing.assert_allclose(global_fit_params[0], M0_VAL, rtol=0.1, atol=0)
    # End of - Fit with fixed DH


def test_fit_tc_unfolding_shared_slopes_many_signals_exponential():

    # p0 includes initial parameters for m, m1 and the shared intercepts
    # The pre-exp and exp coefficients are fitted separately
    p0 = [M0_VAL, M1_VAL, INTERCEPT_N, INTERCEPT_U] + [1,1]*len(temp_list) + [0,0]*len(temp_list) + [ALPHA_N_VAL, ALPHA_U_VAL]*len(temp_list)

    low_bounds = [-1      for _ in p0]
    high_bounds = [np.inf for _ in p0]

    kwargs = {
        'list_of_temperatures' : temp_list,
        'list_of_signals' : signal_list,
        'denaturant_concentrations' : concs,
        'signal_fx' : signal_two_state_tc_unfolding,
        'baseline_native_fx':exponential_baseline,
        'baseline_unfolded_fx':exponential_baseline,
        'fit_m1' : True,
        'tm_value' : Tm_VAL,
        'dh_value' : DHm_VAL,
        'cp_value' : CP0_VAL,
        'signal_ids' : [0 for _ in range(len(signal_list))]
    }

    global_fit_params, cov, predicted_lst = fit_tc_unfolding_shared_slopes_many_signals(
        initial_parameters=p0,
        low_bounds=low_bounds,
        high_bounds=high_bounds,
        **kwargs
    )

    np.testing.assert_allclose(global_fit_params[0], M0_VAL, rtol=0.1, atol=0)


def test_fit_tc_unfolding_many_signals_exponential():

    # p0 includes initial parameters for m, m1 and the shared intercepts
    # The pre-exp and exp coefficients are fitted separately
    p0 = [Tm_VAL,DHm_VAL,M0_VAL,M1_VAL]
    p0 += [INTERCEPT_N, INTERCEPT_U]
    p0 += [ALPHA_N_VAL, ALPHA_U_VAL]
    p0 += [PRE_EXP_N, PRE_EXP_U]
    p0 += [C_N_VAL, C_U_VAL]
    p0 += [ALPHA_N_VAL,ALPHA_U_VAL]

    low_bounds = [-0.1 for _ in p0]
    high_bounds = [1e3 for _ in p0]

    M1_VAL_index = 3
    low_bounds[M1_VAL_index] = -0.05
    high_bounds[M1_VAL_index] = 0.05

    kwargs = {
        'list_of_temperatures' : temp_list,
        'list_of_signals' : signal_list,
        'denaturant_concentrations' : concs,
        'signal_fx' : signal_two_state_tc_unfolding,
        'signal_ids' : [0 for _ in range(len(signal_list))],
        'model_scale_factor': False,
        'cp_value' : CP0_VAL,
        'fit_m1' : True,
        'baseline_native_fx':exponential_baseline,
        'baseline_unfolded_fx':exponential_baseline
    }

    global_fit_params, cov, predicted_lst = fit_tc_unfolding_many_signals(
        initial_parameters=p0,
        low_bounds=low_bounds,
        high_bounds=high_bounds,
        **kwargs
    )

    np.testing.assert_allclose(global_fit_params[:4], [Tm_VAL,DHm_VAL,M0_VAL,M1_VAL], rtol=0.1, atol=1e-2)



"""
This module contains helper functions to obtain the signal, given certain parameters
Author: Osvaldo Burastero
"""

from .rates import (
    eq_constant_termochem,
    eq_constant_thermo
)

from .fractions import (
    fn_two_state_monomer,
    fu_two_state_dimer,
    fu_two_state_trimer,
    fu_two_state_tetramer,
)
import numpy as np

from .math import shift_temperature_K



def signal_two_state_tc_unfolding(
        T,D,DHm,Tm,Cp0,m0,m1,
        p1_N, p2_N, p3_N, p4_N,
        p1_U, p2_U, p3_U, p4_U,
        baseline_N_fx,
        baseline_U_fx,
        extra_arg=None):

    """
    Ref: Louise Hamborg et al., 2020. Global analysis of protein stability by temperature and chemical
    denaturation

    Parameters
    ----------
    T : array-like
        Temperature in Kelvin units
    D : array-like
        Denaturant agent concentration
    DHm : float
        Variation of enthalpy between the two considered states at Tm
    Tm : float
        Temperature at which the equilibrium constant equals one, in Kelvin units
    Cp0 : float
        Variation of calorific capacity between the two states
    m0 : float
        m-value at the reference temperature (Tref)
    m1 : float
        Variation of m-value with temperature
    p1_N, p2_N, p3_N, p4_N : float
        parameters describing the native-state baseline
    p1_U, p2_U, p3_U, p4_U : float
        parameters describing the unfolded-state baseline
    baseline_N_fx : function
        for the native-state baseline
    baseline_U_fx : function
        for the unfolded-state baseline
    extra_arg : None, optional
        Not used but present for API compatibility with oligomeric models
    Returns
    -------
    numpy.ndarray
        Signal at the given temperatures and denaturant agent concentration, given the parameters
    """

    K   = eq_constant_termochem(T,D,DHm,Tm,Cp0,m0,m1)
    fn  = fn_two_state_monomer(K)
    fu  = 1 - fn
    dT   = shift_temperature_K(T)

    # Baseline signals (with quadratic dependence on temperature)
    S_native   = baseline_N_fx(dT,D,p1_N, p2_N, p3_N, p4_N)
    S_unfolded = baseline_U_fx(dT,D,p1_U, p2_U, p3_U, p4_U)

    return  fn*(S_native) + fu*(S_unfolded)


def signal_two_state_t_unfolding(
        T,Tm,dHm,
        p1_N, p2_N, p3_N,
        p1_U, p2_U, p3_U,
        baseline_N_fx,
        baseline_U_fx,
        Cp=0,
        extra_arg=None):

    """
    Two-state temperature unfolding (monomer).

    Parameters
    ----------
    T : array-like
        Temperature
    Tm : float
        Temperature at which the equilibrium constant equals one
    dHm : float
        Variation of enthalpy between the two considered states at Tm
    p1_N, p2_N, p3_N : float
        baseline parameters for the native-state baseline
    p1_U, p2_U, p3_U : float
        baseline parameters for the unfolded-state baseline
    baseline_N_fx : callable
        function to calculate the baseline for the native state
    baseline_U_fx : callable
        function to calculate the baseline for the unfolded state
    Cp : float, optional
        Variation of heat capacity between the two states (default: 0)
    extra_arg : None, optional
        Not used but present for compatibility

    Returns
    -------
    numpy.ndarray
        Signal at the given temperatures, given the parameters
    """

    K   = eq_constant_thermo(T,dHm,Tm,Cp)
    fn  = fn_two_state_monomer(K)
    fu  = 1 - fn

    dT  = shift_temperature_K(T)

    S_native   = baseline_N_fx(dT,0,0,p1_N,p2_N,p3_N) # No denaturant dependence, that's why d=0 and den_slope = 0
    S_unfolded = baseline_U_fx(dT,0,0,p1_U,p2_U,p3_U) # No denaturant dependence, that's why d=0 and den_slope = 0

    return fn*(S_native) + fu*(S_unfolded)

# Oligomeric thermal unfolding signals

def two_state_thermal_unfold_curve(
        T,C,Tm,dHm,
        p1_N, p2_N, p3_N, p4_N,
        p1_U, p2_U, p3_U, p4_U,
        baseline_N_fx,
        baseline_U_fx,
        Cp=0):

    """
    Two-state temperature unfolding (monomer).
    N ⇔ U

    Parameters
    ----------
    T : array-like
        Temperature
    C : array-like
        Oligomer sample concentration
    Tm : float
        Temperature at which the equilibrium constant equals one
    dHm : float
        Variation of enthalpy between the two considered states at Tm
    p1_N, p2_N, p3_N : float
        baseline parameters for the native-state baseline
    p1_U, p2_U, p3_U : float
        baseline parameters for the unfolded-state baseline
    baseline_N_fx : callable
        function to calculate the baseline for the native state
    baseline_U_fx : callable
        function to calculate the baseline for the unfolded state
    Cp : float, optional
        Variation of heat capacity between the two states (default: 0)

    Returns
    -------
    numpy.ndarray
        Signal at the given temperatures, given the parameters
    """

    K   = eq_constant_thermo(T,dHm,Tm,Cp)
    fn  = fn_two_state_monomer(K)
    fu  = 1 - fn

    dT  = shift_temperature_K(T)

    S_native   = baseline_N_fx(dT,0,0,p2_N,p3_N,p4_N)
    S_unfolded = baseline_U_fx(dT,0,0,p2_U,p3_U,p4_U)

    return C * (fn * (S_native) + fu * (S_unfolded))

def two_state_thermal_unfold_curve_dimer(
        T,C,Tm,dHm,
        p1_N, p2_N, p3_N, p4_N,
        p1_U, p2_U, p3_U, p4_U,
        baseline_N_fx,
        baseline_U_fx,
        Cp=0):
    
    """
    Two-state temperature unfolding (dimer).
    N2 ⇔ 2U   C is the total concentration (M) of the protein in dimer equivalent.

    Parameters
    ----------
    T : array-like
        Temperature
    C : array-like
        Oligomer sample concentration
    Tm : float
        Temperature at which the equilibrium constant equals one
    dHm : float
        Variation of enthalpy between the two considered states at Tm
    p1_N, p2_N, p3_N : float
        baseline parameters for the native-state baseline
    p1_U, p2_U, p3_U : float
        baseline parameters for the unfolded-state baseline
    baseline_N_fx : callable
        function to calculate the baseline for the native state
    baseline_U_fx : callable
        function to calculate the baseline for the unfolded state
    Cp : float, optional
        Variation of heat capacity between the two states (default: 0)

    Returns
    -------
    numpy.ndarray
        Signal at the given temperatures, given the parameters

    Notes
    -----
    C is the total concentration (M) of the protein in dimer equivalent.

    """

    K  = eq_constant_thermo(T,dHm,Tm,Cp)
    fu = fu_two_state_dimer(K,C)
    fn = 1 - fu

    dT  = shift_temperature_K(T)

    S_native   = baseline_N_fx(dT,0,0,p2_N,p3_N,p4_N)
    S_unfolded = baseline_U_fx(dT,0,0,p2_U,p3_U,p4_U)

    return C * (fn * (S_native) + fu * (S_unfolded) * 2)

def two_state_thermal_unfold_curve_trimer(
        T,C,Tm,dHm,
        p1_N, p2_N, p3_N, p4_N,
        p1_U, p2_U, p3_U, p4_U,
        baseline_N_fx,
        baseline_U_fx,
        Cp=0):

    """
    Two-state temperature unfolding (trimer).
    N3 ⇔ 3U

    Parameters
    ----------
    T : array-like
        Temperature
    C : array-like
        Oligomer sample concentration
    Tm : float
        Temperature at which the equilibrium constant equals one
    dHm : float
        Variation of enthalpy between the two considered states at Tm
    p1_N, p2_N, p3_N : float
        baseline parameters for the native-state baseline
    p1_U, p2_U, p3_U : float
        baseline parameters for the unfolded-state baseline
    baseline_N_fx : callable
        function to calculate the baseline for the native state
    baseline_U_fx : callable
        function to calculate the baseline for the unfolded state
    Cp : float, optional
        Variation of heat capacity between the two states (default: 0)

    Returns
    -------
    numpy.ndarray
        Signal at the given temperatures, given the parameters

    Notes
    -----
    C is the total concentration (M) of the protein in dimer equivalent.

    """

    K  = eq_constant_thermo(T,dHm,Tm,Cp)
    fu = fu_two_state_trimer(K,C)
    fn = 1 - fu

    dT  = shift_temperature_K(T)

    S_native   = baseline_N_fx(dT,0,0,p2_N,p3_N,p4_N)
    S_unfolded = baseline_U_fx(dT,0,0,p2_U,p3_U,p4_U)

    return C * (fn * (S_native) + fu * (S_unfolded) * 3)

def two_state_thermal_unfold_curve_tetramer(
        T,C,Tm,dHm,
        p1_N, p2_N, p3_N, p4_N,
        p1_U, p2_U, p3_U, p4_U,
        baseline_N_fx,
        baseline_U_fx,
        Cp=0,
        extra_arg=None):

    """
    Two-state temperature unfolding (tetramer).
    N4 ⇔ 4U

    Parameters
    ----------
    T : array-like
        Temperature
    C : array-like
        Oligomer sample concentration
    Tm : float
        Temperature at which the equilibrium constant equals one
    dHm : float
        Variation of enthalpy between the two considered states at Tm
    p1_N, p2_N, p3_N : float
        baseline parameters for the native-state baseline
    p1_U, p2_U, p3_U : float
        baseline parameters for the unfolded-state baseline
    baseline_N_fx : callable
        function to calculate the baseline for the native state
    baseline_U_fx : callable
        function to calculate the baseline for the unfolded state
    Cp : float, optional
        Variation of heat capacity between the two states (default: 0)

    Returns
    -------
    numpy.ndarray
        Signal at the given temperatures, given the parameters

    Notes
    -----
    C is the total concentration (M) of the protein in dimer equivalent.

    """


    K  = eq_constant_thermo(T,dHm,Tm,Cp)
    fu = fu_two_state_tetramer(K,C)
    fn = 1 - fu

    dT  = shift_temperature_K(T)

    S_native   = baseline_N_fx(dT,0,0,p2_N,p3_N,p4_N)
    S_unfolded = baseline_U_fx(dT,0,0,p2_U,p3_U,p4_U)

    return C * (fn * (S_native) + fu * (S_unfolded) * 4)

def map_two_state_model_to_signal_fx(model):
    """

    Maps the model string to the signal type

    Parameters
    ----------
    model : str,
        string representation of model type.

    Returns
    -------
    function
        signal function corresponding to the string

    """
    signal_fx_map = {
    'Monomer':  two_state_thermal_unfold_curve,
    'Dimer':    two_state_thermal_unfold_curve_dimer,
    'Trimer':   two_state_thermal_unfold_curve_trimer,
    'Tetramer': two_state_thermal_unfold_curve_tetramer
    }

    return signal_fx_map.get(model)



"""
This module contains helper functions to parse Differential Scanning Fluorimetry files from different instrument providers
Author: Osvaldo Burastero

All functions that import files should return:
- signal_data_dic: dictionary with the signal data, one entry per signal
- temp_data_dic: dictionary with the temperature data, one entry per signal
- conditions: list with the names of the samples
- signals: list with the names of the signals

A signal can be "350nm", "330nm", "Scattering", "Ratio", "Turbidity", "Ratio 350nm/330nm", etc.
The length of the lists in signal_data_dic and temp_data_dic should be the same as the length of conditions
"""

import numpy  as np
import pandas as pd

import codecs
import json
import re
import os

from openpyxl import load_workbook
from xlrd     import open_workbook

from collections import Counter

from .processing import *

__all__ = [
    "load_csv_file",
    "load_aunty_xlsx",
    "load_quantstudio_txt",
    "load_thermofluor_xlsx",
    "load_nanoDSF_xlsx",
    "load_panta_xlsx",
    "load_uncle_multi_channel",
    "load_mx3005p_txt",
    "detect_file_type",
    "detect_encoding",
    "read_jasco_thermal_ramp"
]

def get_sheet_names_of_xlsx(filepath):
    """
    Get the sheet names of a xls or xlsx file without loading it.
    The open_workbook function is used so we can handle the error "openpyxl does not support the old .xls file format"

    Parameters
    ----------
    filepath : str
        Path to the xls or xlsx file

    Returns
    -------
    list
        List of sheet names
    """

    try:

        wb = load_workbook(filepath, read_only=True, keep_links=False)
        sheetnames = wb.sheetnames

    except:

        xls = open_workbook(filepath, on_demand=True)
        sheetnames = xls.sheet_names()

    return sheetnames

def file_is_of_type_aunty(file_path):

    """
    Detect if file is an AUNTY xlsx file.

    The AUNTY format contains multiple sheets where the first column is
    temperatures and subsequent columns are fluorescence values. The first
    row contains the word 'wavelength' and the second row contains the word
    'temperature' in the first column.

    Parameters
    ----------
        file_path (str): Path to the .xls or .xlsx file to test.

    Returns
    -------
        bool: True if the file matches the AUNTY format heuristic, False otherwise.
    """

    if not (file_path.endswith('.xls') or file_path.endswith('.xlsx')):
        return False

    sheet_names = get_sheet_names_of_xlsx(file_path)

    for sheet_name in sheet_names:

        # Load the data
        data = pd.read_excel(
            file_path, sheet_name=sheet_name,
            header=None,skiprows=0)

        try:

            wavelength_cell = data.iloc[0, 1]
            temperature_cell = data.iloc[1, 0]
            corner_cell = data.iloc[0, 0]

            # Verify that the word wavelength is in the first row, second column
            condition1 = isinstance(wavelength_cell, str) and 'wavelength' in wavelength_cell.lower()

            # Verify that the word temperature is in the second row, first column
            condition2 = isinstance(temperature_cell, str) and 'temperature' in temperature_cell.lower()

            # Verify that the corner cell is empty (NaN)
            condition3 = np.isnan(corner_cell)

            if not (condition1 and condition2 and condition3):
                continue

            return True

        except:

            file_is_aunty = False

    return file_is_aunty

def file_is_of_type_uncle(xlsx_file):

    """
    Check if the file is an uncle file

    Parameters
    ----------
    xlsx_file : str
        Path to the xlsx file

    Returns
    -------
    bool
        True if the file is an uncle file, False otherwise
    """

    try:

        # Read the first sheet of the xlsx file
        data = pd.read_excel(xlsx_file,skiprows=1,nrows=5,header=None)

        # Extrac the first row
        row = data.iloc[0,1:]

        # concatenate the row into a string
        row_str = ' '.join([str(x) for x in row])

        # count the number of times we have word 'Time' and 'Temp'
        count_time = row_str.count('Time')
        count_temp = row_str.count('Temp')

        if count_time > 20 and count_temp > 20 and count_temp == count_time:
            return True
        else:
            return False
    except:
        return False

def detect_file_type(file):

    """
    Detect the type of file based on its extension and content.

    Parameters
    ----------
    file : str
        Path to the file

    Returns
    -------
    str or None
        Type of file (e.g., 'supr', 'csv', 'prometheus', 'panta', 'uncle', 'mx3005p', 'quantstudio', etc.) or None if unknown
    """

    file_extension = file.split('.')[-1]

    if file_extension == 'supr':
        return 'supr'

    if file_extension in ["xlsx", "xls"]:
        # Get file type: DSF or nDSF
        sheet_names = get_sheet_names_of_xlsx(file)

        # Load the data to the Python class
        if "RFU" in sheet_names:
            return 'thermofluor'
        elif "Data Export" in sheet_names or "melting-scan" in sheet_names:
            return 'panta'
        elif file_is_of_type_uncle(file):
            return 'uncle'
        elif file_is_of_type_aunty(file):
            return 'aunty'
        else:
            return 'prometheus'

    if file_is_jasco_thermal_ramp(file):
        return "jasco_thermal_ramp"

    if file_is_jasco_thermal_ramp_format_2(file):
        return "jasco_thermal_ramp_format_2"

    if file_extension == 'csv':

        if file_is_chirascan_thermal_ramp(file):

            return "chirascan_thermal_ramp"

        else:

            return 'csv'

    with codecs.open(file, 'r', encoding='utf-8',errors='ignore') as rf:
        ls       = rf.read().splitlines()

        for line in ls:
            if line.startswith('Segment') and 'Well' in line:
                return 'mx3005p'

            splittedLine = line.split()
            if 'Well' in splittedLine and 'Target' in splittedLine and 'Reading' in splittedLine:
                return 'quantstudio'

    # Return an error if no extension is found
    raise ValueError(f'File extension not recognized: {file_extension}')

def load_aunty_xlsx(file_path):

    """
    Load AUNTY-format multi-sheet Excel file where each sheet is a condition.

    Parameters
    ----------
    file_path : str
        Path to the AUNTY xlsx file
    """
    # Get the names of the sheets
    sheet_names = get_sheet_names_of_xlsx(file_path)

    wavelengths       = None
    temperatures      = []

    conditions        = []
    signals_data      = []

    signal_data_dic = {}
    temp_data_dic   = {}

    for sheet_name in sheet_names:

        # Load the data
        data = pd.read_excel(
            file_path, sheet_name=sheet_name,
            header=None,skiprows=0
            )

        wavelength_cell = data.iloc[0, 1]
        temperature_cell = data.iloc[1, 0]
        corner_cell = str(data.iloc[0, 0])

        # Verify that the word wavelength is in the first row, second column
        condition1 = isinstance(wavelength_cell, str) and 'wavelength' in wavelength_cell.lower()

        # Verify that the word temperature is in the second row, first column
        condition2 = isinstance(temperature_cell, str) and 'temperature' in temperature_cell.lower()
        
        # Verify that the corner cell is empty (NaN)
        condition3 = corner_cell == '' or corner_cell.lower() == 'nan'

        if not (condition1 and condition2 and condition3):
            continue

        signal_data  = np.array(data.iloc[2:, 1:]).astype(float)

        if wavelengths is None:

            wavelengths    = np.round(np.array(data.iloc[1, 1:]).astype(float),2)

        temperature_data   = np.round(np.array(data.iloc[2:, 0]).astype(float), 2)

        signals_data.append(signal_data)
        temperatures.append(temperature_data)
        conditions.append(sheet_name)


    # Now we have one signal matrix per condition, but we need one signal matrix per wavelength, with
    # the conditions as columns
    # Therefore, for each wavelength, we create one dataframe per condition and then merge them

    named_wls = [str(wl) + ' nm' for wl in wavelengths]

    for i, named_wl in enumerate(named_wls):

        signal_data_as_list = [arr[:,i] for arr in signals_data]

        signal_data_dic[named_wl] = signal_data_as_list
        temp_data_dic[named_wl]   = temperatures

    return signal_data_dic, temp_data_dic, conditions, named_wls   

def detect_encoding(file_path):

    """
    Detect the encoding of a file by trying common encodings.

    Parameters
    ----------
    file_path : str
        Path to the file

    Returns
    -------
    str
        Detected encoding or the string 'Unknown encoding'
    """

    try:
        with codecs.open(file_path, encoding="utf-8", errors="strict") as f:
            f.read()
        return 'utf-8'
    
    except:
        with codecs.open(file_path, encoding="latin1", errors="strict") as f:
            f.read()
        return 'latin1'

def find_indexes_of_non_signal_conditions(signal_data,conditions):

    """
    Given the signal data and the conditions, find the indexes of the conditions that are not signal data.
    We assume that conditions with the word 'derivative' or 'S.D.' in it are not signal data.

    Parameters
    ----------
    signal_data : numpy.ndarray
        2D array with the signal data (shape: temperature x conditions)
    conditions : list
        List of condition names

    Returns
    -------
    list
        List of indexes to remove
    """

    # Find the indexes of conditions with derivative in it, or 'S.D.' in it
    idx_to_remove1 = [i for i, cond in enumerate(conditions) if 'derivative' in cond.lower() or 's.d.' in cond.lower()]

    # Find indexes empty signal_data columns
    idx_to_remove2 = [i for i in range(signal_data.shape[1]) if np.isnan(signal_data[:, i]).all()]

    # Find columns where all values are NaN
    nan_columns = np.all(np.isnan(signal_data), axis=0)

    # Get the column indices
    idx_to_remove3 = np.where(nan_columns)[0].tolist()

    # Find unique indexes to remove
    idx_to_remove = list(set(idx_to_remove1 + idx_to_remove2 + idx_to_remove3))

    return idx_to_remove

def find_repeated_words(string_lst):

    """
    Given a list of strings, find the repeated words in the list.

    Parameters
    ----------
    string_lst : list of str
        List of strings to analyze

    Returns
    -------
    list
        List of repeated words
    """

    # Repeated words
    words = " ".join(string_lst).split()

    # Count the occurrences of each word
    word_counts = Counter(words)

    # Find words that are repeated (appear, at least, as many times as elements in string_lst)
    repeated_words = [word for word, count in word_counts.items() if count >= len(string_lst)]

    return repeated_words

def remove_words_in_string(input_string,word_list):

    """
    Given a string and a list of words, remove the words from the string.

    Parameters
    ----------
    input_string : str
        String to be processed
    word_list : list of str
        List of words to remove

    Returns
    -------
    str
        Processed string
    """

    # Split the string into words
    words = input_string.split()

    # Replace words from the list with an empty string
    filtered_words = [word for word in words if word not in word_list]

    # Join the words back into a single string
    output_string = ' '.join(filtered_words)

    return output_string

def load_csv_file(file):

    """
    Load a CSV file containing temperature and signal columns and return structured data.

    Parameters
    ----------
    file : str
        Path to the csv file

    Returns
    -------
    signal_data_dic : dict
        Dictionary mapping signal names to lists of 1D numpy arrays (one array per condition)
    temp_data_dic : dict
        Dictionary mapping signal names to lists of temperature arrays corresponding to the signals
    conditions : list
        List of condition names
    signals : numpy.ndarray
        Array of signal name strings
    """

    signal_data_dic = {}
    temp_data_dic   = {}

    signals = []

    encoding = detect_encoding(file)

    # Try common delimiters
    for delimiter in [',', ';', '\t']:

        try:

            dat = pd.read_csv(file, delimiter=delimiter, encoding=encoding)

            # Convert non-numeric columns to NaN
            dat = dat.apply(pd.to_numeric, errors='coerce')

            # Produce error if we don't have 2 or more columns
            if len(dat.columns) < 2:
                raise ValueError('File does not have enough columns')

            # Set the conditions names and start index for the signal data
            if 'time' in dat.columns[0].lower():
                conditions = [str(c) for c in dat.columns[2:]]
                idx_start = 1
            else:
                conditions = [str(c) for c in dat.columns[1:]]
                idx_start = 0

            signal_data = np.array(dat.iloc[:, (idx_start + 1):]).astype('float')
            temperature_data = np.array(dat.iloc[:, idx_start]).astype('float')

            break

        except:

            pass

    idx_to_remove = find_indexes_of_non_signal_conditions(signal_data, conditions)

    # Remove the elements from the conditions array
    conditions = [cond for i, cond in enumerate(conditions) if i not in idx_to_remove]

    # Remove the columns from the signal data
    signal_data = np.delete(signal_data, idx_to_remove, axis=1)

    # Divide the conditions into groups, according to the presence of the words '350', '330', 'ratio', 'scattering'
    conditions_350nm = [cond for cond in conditions if '350' in cond.lower() and 'ratio' not in cond.lower()]
    conditions_330nm = [cond for cond in conditions if '330' in cond.lower() and 'ratio' not in cond.lower()]
    conditions_ratio = [cond for cond in conditions if 'ratio' in cond.lower()]
    conditions_scatt = [cond for cond in conditions if 'scattering' in cond.lower()]

    # Check that the number of conditions is the same for all groups
    conditions_lst = [conditions_350nm, conditions_330nm, conditions_ratio, conditions_scatt]
    sel_cond_lst   = [cond for cond in conditions_lst if len(cond) > 0]
    n_conditions   = [len(cond) for cond in conditions_lst if len(cond) > 0]

    possible_rep_conditions = len(np.unique(n_conditions)) == 1

    # case where the CSV corresponds to a nDSF csv
    if possible_rep_conditions:

        # For each group, store the signal data and the temperature data
        if len(conditions_350nm) > 0:
            signal_350 = signal_data[:, [i for i, cond in enumerate(conditions) if cond in conditions_350nm]]
            signal_data_as_list = signal_350.transpose().tolist()
            signal_data_dic["350 nm"] = signal_data_as_list
            temp_data_dic["350 nm"] = expand_temperature_list([temperature_data],signal_data_as_list)
            signals.append("350 nm")

        if len(conditions_330nm) > 0:
            signal_330 = signal_data[:, [i for i, cond in enumerate(conditions) if cond in conditions_330nm]]
            signal_data_as_list = signal_330.transpose().tolist()
            signal_data_dic["330 nm"] = signal_data_as_list
            temp_data_dic["330 nm"] = expand_temperature_list([temperature_data],signal_data_as_list)
            signals.append("330 nm")

        if len(conditions_ratio) > 0:
            signal_ratio = signal_data[:, [i for i, cond in enumerate(conditions) if cond in conditions_ratio]]
            signal_data_as_list = signal_ratio.transpose().tolist()
            signal_data_dic["Ratio 350 nm / 330 nm"] = signal_data_as_list
            temp_data_dic["Ratio 350 nm / 330 nm"] = expand_temperature_list([temperature_data],signal_data_as_list)
            signals.append("Ratio 350 nm / 330 nm")

        if len(conditions_scatt) > 0:
            signal_scatt = signal_data[:, [i for i, cond in enumerate(conditions) if cond in conditions_scatt]]
            signal_data_as_list = signal_scatt.transpose().tolist()
            signal_data_dic["Scattering"] = signal_data_as_list
            temp_data_dic["Scattering"] = expand_temperature_list([temperature_data],signal_data_as_list)
            signals.append("Scattering")

        cond_temp = sel_cond_lst[0]

        if len(cond_temp) > 1:

            repeated_words = find_repeated_words(cond_temp)
            cond_temp = [remove_words_in_string(cond, repeated_words) for cond in cond_temp]
            conditions = cond_temp

        else:

            conditions = cond_temp

    else:

        # If we only have one condition, use the condition as signal name
        if len(conditions) == 1:
            signal_name = conditions[0]
        # Default signal name
        else:
            signal_name = "Fluorescence"

        signal_data_as_list = signal_data.transpose().tolist()
        signal_data_dic[signal_name] = signal_data_as_list
        temp_data_dic[signal_name] = expand_temperature_list([temperature_data],signal_data_as_list)
        signals.append(signal_name)

    signals = np.array(signals)

    return signal_data_dic, temp_data_dic, conditions, signals

def get_start_line_quantstudio_txt(file_name):

    """
    Find the start line for QuantStudio text files (first non-comment data line).

    Parameters
    ----------
    file_name : str
        Path to the QuantStudio txt file

    Returns
    -------
    int
        Number  of the first line not starting with '*' plus one
    """

    with codecs.open(file_name, 'r', encoding='utf-8',errors='ignore') as rf:
        ls       = rf.read().splitlines()
        splitted = [l.split() for l in ls]
        
        for i,s in enumerate(splitted):
            if len(s) > 5 and "*" not in s[0]:
                start_row = i+1
                break

    return(start_row)

def load_quantstudio_txt(QSfile):

    """
    Load QuantStudio TXT files (.txt) exported from QuantStudio instruments.

    Parameters
    ----------
    QSfile : str
        Path to the QuantStudio txt file

    Returns
    -------
    signal_data_dic : dict
        Dictionary with signal data (key: 'Fluorescence')
    temp_data_dic : dict
        Dictionary with temperature arrays per condition
    conditions : list
        List of condition names (well identifiers)
    signals : numpy.ndarray
        Array with signal name(s)
    """

    start_row = get_start_line_quantstudio_txt(QSfile)
    data      = pd.read_csv(QSfile,skiprows=start_row,sep=r"\s+",header=None)

    signal_data_dic = {'Fluorescence': []}
    temp_data_dic   = {'Fluorescence': []}

    u, ind     = np.unique(data.iloc[:,1], return_index=True)
    conditions = u[np.argsort(ind)].tolist()

    signal = "Fluorescence"

    # Process each group directly without creating DataFrames
    for _, group in data.groupby(1):
        
        # Extract temperature and signal columns (columns 3 and 4, 0-indexed)
        temp_data = group.iloc[:, 3].astype(str).str.replace(',', '').astype(float).values
        signal_data = group.iloc[:, 4].astype(str).str.replace(',', '').astype(float).values
        
        # Sort by temperature
        sorted_indices = np.argsort(temp_data)
        temp_data = temp_data[sorted_indices]
        signal_data = signal_data[sorted_indices]

        signal_data_dic[signal].append(signal_data)
        temp_data_dic[signal].append(temp_data)

    signals = np.array([signal])

    return signal_data_dic, temp_data_dic, conditions, signals

def load_thermofluor_xlsx(thermofluor_file):

    """
    Load DSF Thermofluor xls file and extract data.

    Parameters
    ----------
    thermofluor_file : str
        Path to the xls file

    Returns
    -------
    signal_data_dic : dict
        Dictionary with signal data
    temp_data_dic : dict
        Dictionary with temperature data
    conditions : list
        List of conditions
    """

    xls  = pd.ExcelFile(thermofluor_file)
    dat = pd.read_excel(xls, "RFU",header=None)
    conditions = np.array(dat.iloc[0, 1:]).tolist()

    signal_data_dic = {}
    temp_data_dic   = {}

    signal = "DSF_RFU"

    fluo = np.array(dat.iloc[1:,1:]).astype('float').transpose().tolist()
    temp = [np.array(dat.iloc[1:, 0]).astype('float')]

    signal_data_dic[signal] = fluo
    temp_data_dic[signal] = expand_temperature_list(temp, fluo)

    signals = np.array([signal],dtype='str')

    return signal_data_dic, temp_data_dic, conditions, signals

def space_combinations(word):
    """
    Generate all combinations of a word with 0, 1, or 2 spaces before and/or after the word.

    Parameters
    ----------
    word : str
        The word to generate combinations for.

    Returns
    -------
    list
        List of strings with all combinations of spaces before and after the word.
    """
    spaces = [' ', '  ', '   ']
    combos = []
    for before in ['', *spaces]:
        for after in ['', *spaces]:
            combos.append(f"{before}{word}{after}")
    return combos

def load_nanoDSF_xlsx(processed_dsf_file):

    """
    Load nanotemper processed xlsx file and extract relevant data.

    Parameters
    ----------
    processed_dsf_file : str
        Path to the processed xlsx file

    Returns
    -------
    signal_data_dic : dict
        Dictionary with signal data
    temp_data_dic : dict
        Dictionary with temperature data
    conditions : list
        List of conditions
    signals : numpy.ndarray
        Array of signal names
    """

    sheet_names = get_sheet_names_of_xlsx(processed_dsf_file)

    signal_data_dic = {}
    temp_data_dic   = {}

    # this needs to be the processed file!
    xls = pd.ExcelFile(processed_dsf_file)

    conditions_df = pd.read_excel(xls, "Overview")
    conditions = conditions_df['Sample ID'].fillna('').astype(str).tolist()

    # Add position index to avoid problems with empty names

    possible_signals = space_combinations('350nm') + space_combinations('330nm') + space_combinations('Ratio') + space_combinations('Scattering')

    include = []

    # Change signal name if unfolding curve is present
    for sn in sheet_names:
        include_value = any(
            [ps in sn and "deriv" not in sn.lower() and "fold" not in sn.lower() for ps in possible_signals])

        include.append(include_value)

    sheet_names_to_load = np.array([s for (i, s) in zip(include, sheet_names) if i])

    signals = np.array([" ".join(sn.split()) for sn in sheet_names_to_load])

    for sn, signal in zip(sheet_names_to_load, signals):

        dat = pd.read_excel(xls, sn, index_col=None, header=None)

        indices   = np.argwhere(dat.iloc[:, 0].values == 'Time [s]')
        first_row = int(indices[0][0]) + 1

        # Find all the columns with the word temperature
        column_headers = dat.iloc[first_row - 1, :].values

        ids_temperature = [i for i, x in enumerate(column_headers) if 'temperature' in x.lower()]
        ids_time        = [i for i, x in enumerate(column_headers) if 'time'        in x.lower()]

        # If we have more than one temperature column, remove the extra temperature and time columns
        if len(ids_temperature) > 1:
            ids_temperature = ids_temperature[1:]  # We keep the first temperature colummn
            ids_time        = ids_time[1:]  # We keep the first time column

            # Remove columns of dataframe by index
            dat = dat.drop(dat.columns[ids_temperature + ids_time], axis=1)

        fluo = np.array(dat.iloc[first_row:, 2:]).astype('float').transpose().tolist()
        temp = [np.array(dat.iloc[first_row:, 1]).astype('float')]

        signal_data_dic[signal] = fluo
        temp_data_dic[signal]   = expand_temperature_list(temp,fluo)

    return signal_data_dic, temp_data_dic, conditions, signals

def load_panta_xlsx(pantaFile):

    """
    Load the xlsx file generated by a Prometheus Panta instrument.

    Parameters
    ----------
    pantaFile : str
        Path to the xlsx file

    Returns
    -------
    signal_data_dic : dict
        Dictionary with signal data
    temp_data_dic : dict
        Dictionary with temperature data
    conditions : list
        List of conditions
    signals : numpy.ndarray
        List of signal names, such as 330nm and 350nm
    """

    sheet_names = get_sheet_names_of_xlsx(pantaFile)

    data_sheetname = "Data Export" if "Data Export" in sheet_names else "melting-scan"

    data = pd.read_excel(pantaFile, data_sheetname)

    signal_data_dic = {}
    temp_data_dic   = {}

    column_names  = [str.lower(c) for c in data.columns]

    pos_350       = [i for i,x in enumerate(column_names) if "350"        in x and "deriv" not in x and "330" not in x]
    pos_330       = [i for i,x in enumerate(column_names) if "330"        in x and "deriv" not in x and "350" not in x]
    scattering    = [i for i,x in enumerate(column_names) if "scattering" in x and "deriv" not in x]
    pos_ratio     = [i for i,x in enumerate(column_names) if "ratio"      in x and "deriv" not in x]
    pos_turb      = [i for i,x in enumerate(column_names) if "turbidity"  in x and "deriv" not in x]

    possible_signals    = ["350 nm","330 nm","Scattering","Ratio","Turbidity"]
    signals             = []

    all_positions = [pos_350,pos_330,scattering,pos_ratio,pos_turb]

    for positions,signal in zip(all_positions,possible_signals):

        signal_data = []
        temp_data   = []

        if len(positions) > 0:

            for position in positions:

                # Find the index of the previous column with the word temperature
                temp_col = [i for i, x in enumerate(column_names) if 'temperature' in x.lower() and i < position][-1]

                # If the temperature column has the word "refolding", exclude it
                if "refolding" in column_names[temp_col]:
                    continue

                # round to two digits the temperature values
                temperature_values = data.iloc[:, temp_col].to_numpy().astype('float')
                temperature_values = np.round(temperature_values, 3)

                signal_values = data.iloc[:, position].to_numpy().astype('float')

                # Sort the temperature values and signal values together, according to the temperature values
                sorted_indices     = np.argsort(temperature_values)
                temperature_values = temperature_values[sorted_indices]
                signal_values      = signal_values[sorted_indices]

                # Extract the temperature data and signal data
                temp_data.append(temperature_values)
                signal_data.append(signal_values)

            signal_data_dic[signal]   = signal_data
            temp_data_dic[signal]     = temp_data

            signals.append(signal)

    signals        = np.array(signals)

    try:

        conditions_df = pd.read_excel(pantaFile, "Overview")
        conditions    = conditions_df[['Sample ID']].values.flatten().tolist()

    except:

        conditions = np.repeat('',len(signal_data)).tolist()

    return signal_data_dic, temp_data_dic, conditions, signals

def load_uncle_multi_channel(uncle_file):

    """
    Function to load the data from the UNCLE instrument.

    Parameters
    ----------
    uncle_file : str
        Path to the xlsx file

    Returns
    -------
    signal_data_dic : dict
        Dictionary with signal data (keys: wavelength strings like '350 nm')
    temp_data_dic : dict
        Dictionary with temperature arrays per condition
    conditions : list
        List of sample names
    signals : list
        List of wavelength strings
    """

    # Get the names of the sheets
    sheet_names = get_sheet_names_of_xlsx(uncle_file)

    wavelengths       = None
    temperatures      = []

    conditions        = []
    signals_data      = []

    signal_data_dic = {}
    temp_data_dic   = {}

    # Loop through each sheet and read the data
    for sheet_name in sheet_names:

        try:

            # Read the data from the sheet
            data = pd.read_excel(
                uncle_file, 
                sheet_name=sheet_name,
                header=None,
                skiprows=0
                )

            # Extract the sample name, from the first row, fitfh column
            sample_name = data.iloc[0, 4]

            # Remove the first row
            data = data.iloc[1:, :]

            # Extract the time/temperature data
            temperature_data = data.iloc[0, 1:].values

            # Select the temperature data
            temperature_data = [x.split(',')[0] for x in temperature_data]
            temperature_data = [x.split(':')[1] for x in temperature_data]
            temperature_data = np.array(temperature_data,dtype=float)

            # Extract the signal data
            # It contains one column per temperature and one row per wavelength
            signal_data = np.array(
                data.iloc[3:, 1:].values,
                dtype=float
                )

            # Assign the wavelength data if wavelengths is None
            if wavelengths is None:
                wavelengths = np.round(
                    np.array(
                        data.iloc[3:, 0].values,
                        dtype=float
                        ),
                    decimals=1
                    )

            signals_data.append(signal_data) # List of 2D arrays, one per condition
            temperatures.append(temperature_data) # List of 1D arrays, one per condition
            conditions.append(sample_name)

        except:

            pass

    # Iterate over the wavelengths
    named_wls = [str(wl) + ' nm' for wl in wavelengths]

    # We require one signal matrix per wavelength
    # with one column per condition
    for i in range(len(wavelengths)):

        # Extract the signal data for the current wavelength
        # signal_temp has one row per condition and one column per wavelength
        signal_data_i = [arr[i,:] for arr in signals_data]

        wl = named_wls[i]

        signal_data_dic[wl] = signal_data_i
        temp_data_dic[wl]   = temperatures

    return signal_data_dic, temp_data_dic, conditions, named_wls

def load_mx3005p_txt(filename):
    """
    Load Agilent MX3005P qPCR txt file and extract data

    Parameters
    ----------
    filename : str
        Path to the MX3005P txt file. The second column has the fluorescence data, and
        the third column the temperature. Wells are separated by rows containing a sentence like this one: 'Segment  2 Plateau  1 Well  1'

    Returns
    -------
    signal_data_dic : dict
        Dictionary with signal data
    temp_data_dic : dict
        Dictionary with temperature data
    conditions : list
        List of conditions (well numbers)
    signals : numpy.ndarray
        List of signal names
    """
    
    signal_data_dic = {'Fluorescence': []}
    temp_data_dic = {'Fluorescence': []}
    conditions = []
    
    with open(filename, 'r') as f:
        ls = f.read().splitlines()

        for i, line in enumerate(ls):
            if line.startswith('Segment') and 'Well' in line:
                # Get the well number
                well_num = line.split()[-1]
                conditions.append(well_num)

                fluorescence = []
                temperature = []

                # Read data for this well
                for line2 in ls[i+2:]:
                    if line2.startswith('Segment'):
                        break
                    else:
                        data = line2.split()
                        if len(data) >= 3:
                            fluorescence.append(float(data[1]))
                            temperature.append(float(data[2]))

                # Convert to arrays and sort by temperature
                temp_array = np.array(temperature)
                fluo_array = np.array(fluorescence)
                
                sorted_indices = np.argsort(temp_array)
                temp_array = temp_array[sorted_indices]
                fluo_array = fluo_array[sorted_indices]
                
                signal_data_dic['Fluorescence'].append(fluo_array)
                temp_data_dic['Fluorescence'].append(temp_array)

    signals = np.array(['Fluorescence'])
    
    return signal_data_dic, temp_data_dic, conditions, signals

def load_supr_dsf( JSON_file):

    signal_data_dic = {}
    temp_data_dic   = {}

    # Read JSON data from a file
    with open(JSON_file, "r") as file:
        json_data = file.read()

    # Parse JSON data into a dictionary
    data_dict = json.loads(json_data)

    samples_name = [item["SampleName"] for item in data_dict['Samples']]
    samples_well = [item["WellLocations"] for item in data_dict['Samples']]

    samples_name_simple = []
    samples_well_simple = []

    for sn, sw in zip(samples_name, samples_well):

        if ',' in sw:

            sw = sw.split(',')
            sn = [sn for _ in sw]

        else:

            sw = [sw]
            sn = [sn]

        samples_name_simple += sn
        samples_well_simple += sw

    name_df = pd.DataFrame({
        'well': samples_well_simple,
        'name': samples_name_simple})

    scans = [item["_scans"] for item in data_dict['Wells']]
    n_scans = len(scans)

    wavelengths = data_dict['Wavelengths']
    wavelengths = np.round(wavelengths, decimals=1)
    n_wavelengths = len(wavelengths)

    temperatures = []
    signals = []

    well = [item["PhysicalLocation"] for item in data_dict['Wells']]

    # Create a categorical data type with the custom order
    cat_type = pd.CategoricalDtype(categories=well, ordered=True)

    # Convert the column to the categorical data type
    name_df['well'] = name_df['well'].astype(cat_type)

    # Sort the DataFrame based on the custom order
    name_df = name_df.sort_values(by='well')

    conditions = name_df['name'].values.astype(str).tolist()

    for i in range(n_scans):
        temperatures.append([item['Temperature'] for item in scans[i]])
        signals.append([item['Signal'] for item in scans[i]])

    temperatures = np.array(temperatures).T
    temperatures = np.round(temperatures, decimals=1)

    signals = np.array(signals)

    # Iterate over the wavelengths

    named_wls = [str(wl) + ' nm' for wl in wavelengths]

    for i in range(n_wavelengths):

        signals_temp = signals[:, :, i].T

        signals_data = []
        temps_data   = []

        # Iterate over the columns of the arrays
        for ii in range(n_scans):

            x = temperatures[:, ii]
            y = signals_temp[:, ii]

            sorted_indices = np.argsort(x)
            x = x[sorted_indices]
            y = y[sorted_indices]

            # Find np.nans in y and x, and remove them
            non_nas_y = np.logical_not(np.isnan(y))
            non_nas_x = np.logical_not(np.isnan(x))

            non_nas_yx = np.logical_and(non_nas_y, non_nas_x)

            x = x[non_nas_yx]
            y = y[non_nas_yx]

            signals_data.append(y)
            temps_data.append(x)

        wl = named_wls[i]

        signal_data_dic[wl] = signals_data
        temp_data_dic[wl] = temps_data

    return signal_data_dic, temp_data_dic, conditions, named_wls

def find_delimiter_character(string):

    """
    Guess if a string contains ',' separated values or ';' separated values

    Parameters
    ----------
    string : str
        The string to analyze

    Returns
    -------
    delimiter : str
        The delimiter character that is most likely used in the string (',' or ';' or '\t')
    """

    comma_count     = string.count(",")
    semicolon_count = string.count(";")
    tab_count       = string.count("\t")

    counts = {
        ",": comma_count,
        ";": semicolon_count,
        "\t": tab_count
    }

    delimiter = max(counts, key=counts.get)

    # Choose the delimiter with the highest count
    return delimiter


def read_jasco_thermal_ramp(file):

    """
    Given a JASCO file with a thermal ramp, this function reads the data

    The data is given in chuncks:

    Channel 1
	4.94	14.93	25.08	35.02	45.03	55.07	64.99	75.04	85.08	95.03
    250	-0.310564	-0.112003	0.0199744	-0.217282	-0.238716	-0.173046	0.00129784	-0.394731	-0.687165	-1.40543

    Parameters
    ----------
    file : str
        Path to the JASCO thermal ramp file

    Returns
    -------
    signal_data_dic : dict
        Dictionary with signal data (key: wavelength string like '250 nm')
    temp_data_dic : dict
        Dictionary with temperature arrays per condition (only one condition in this case)
    conditions : list
        File name as condition
    wavelength_data : list
        List of wavelength strings (e.g., '250 nm', '255 nm', etc.)

    """

    with open(file, encoding="latin-1") as f:

        ls = f.read().splitlines()

        split_str = find_delimiter_character("".join(ls[:20]))

        channels_idx =[i+15 for i, l in enumerate(ls[15:]) if "Channel" in l]

        # Find the first line with the word 'Channel'
        channel_idx1 = channels_idx[0]

        channel_idx2 = channels_idx[1] if len(channels_idx) > 1 else len(ls)

        # Find the number of temperature points (Line after channel 1)
        temperatures = ls[channel_idx1 + 1].split(split_str)

        # Remove empty lines
        temperatures = [parse_number(x) for x in temperatures if x != '']

        # Convert temperatures to numpy array
        temperatures = np.array(temperatures)

        # Find the wavelength data, first column
        wavelength_data = [str(parse_number(x.split(split_str)[0])) + " nm" for x in ls[(channel_idx1+2):channel_idx2]]

        # Convert wavelength_data to  numpy array
        wavelength_data = np.array(wavelength_data)

        # Retrieve the CD data
        cd_data = [x.split(split_str)[1:] for x in ls[(channel_idx1+2):channel_idx2]]

        for i in range (len(cd_data)):

            cd_data[i] = [parse_number(x) for x in cd_data[i] if x != '']

        signal_data_dic ={}
        temp_data_dic  = {}

        for cd_signal,wl in zip(cd_data,wavelength_data):

            signal_data_dic[wl] = [cd_signal]
            temp_data_dic[wl] = [temperatures]

        # extract file basename
        bname = os.path.basename(file)

        conditions = [bname]

    return signal_data_dic, temp_data_dic, conditions, wavelength_data


def file_is_jasco_thermal_ramp(file):

    """
    Find if a file is of type JASCO thermal ramp

    Parameters
    ----------
    file : str
        Path to the file

    Returns
    -------
    bool
        True if the file is a JASCO thermal ramp, False otherwise

    """

    with open(file, encoding="latin-1") as f:

        ls = f.read().splitlines()

        split_str = find_delimiter_character("".join(ls[:20]))

        condition0 = False

        # Detect if we have a thermal ramp by finding
        # the presence of many numeric columns
        for l in ls[20:50]:

            l_split     = l.split(split_str)

            if are_all_strings_numeric(l_split) and len(l_split) > 4:
                condition0 = True
                break

        l_split = [l.split(split_str) for l in ls[:20]]

        condition1 = any([x[0].lower() == "origin" and x[1].lower() == "jasco"      for x in l_split])
        condition2 = any([x[0].lower() == "xunits" and x[1].lower() == "nanometers" for x in l_split])

        if all([condition0, condition1, condition2]):
            return True

    return False

def file_is_jasco_thermal_ramp_format_2(file):

    """
    Find if a file is of type JASCO thermal ramp, format 2

    Parameters
    ----------
    file : str
        Path to the file

    Returns
    -------
    bool
        True if the file is a JASCO thermal ramp, format 2, False otherwise

    """

    with open(file, encoding="latin-1") as f:

        ls = f.read().splitlines()

        split_str = find_delimiter_character("".join(ls[:20]))

        condition0 = False

        # Detect if we have a thermal ramp by finding
        # the presence of numeric columns, but less than 10
        for l in ls[20:50]:

            l_split     = l.split(split_str)

            if are_all_strings_numeric(l_split) and len(l_split) >= 2 and len(l_split) < 10:
                condition0 = True
                break

        l_split = [l.split(split_str) for l in ls[:20]]

        condition1 = any([x[0].lower() == "origin" and "jasco" in x[1].lower()       for x in l_split])
        condition2 = any([x[0].lower() == "xunits" and "temperature" in x[1].lower() for x in l_split])

        if all([condition0, condition1, condition2]):
            return True

    return False

def read_jasco_thermal_ramp_format_2(file):

    """
    Given a JASCO file with a thermal ramp at only one wavelength, this function reads the data

    XYDATA
    4,0000	-29,1176	295,223
    6,0600	-29,3217	295,577
    8,0600	-29,0102	295,484
    10,0500	-28,9184	295,437
    12,0600	-28,8087	295,167
    14,1000	-28,5928	295,186


    """

    with open(file, encoding="latin-1") as f:

        ls = f.read().splitlines()

        split_str = find_delimiter_character("".join(ls[:20]))

        # Find the wavelength
        line_with_wl_info = [ l for l in ls if "wavelength" in l ][0]

        # Remove all non-numeric characters
        wavelength_value = re.sub("[^0-9,.-]", "", line_with_wl_info)

        # Find the first line with the word 'XYDATA'
        idx_start = [ i for i, l in enumerate(ls) if "XYDATA" in l ][0] + 1

        # Find the first empty line or starting with asterisk after idx_start
        idx_end = [i for i,l in enumerate(ls) if l == '' or l.startswith('*') if i > idx_start][0]

        sel_lines = ls[idx_start:idx_end]

        sel_lines_split = [ l.split(split_str) for l in sel_lines ]

        temperatures = [parse_number(l[0]) for l in sel_lines_split]
        temperatures = np.array(temperatures)

        cd_data = [parse_number(l[1]) for l in sel_lines_split]

    signals = np.array([str(parse_number(wavelength_value)) + " nm"])

    # extract file basename
    bname = os.path.basename(file)

    conditions = [bname]

    signal_data_dic = {signals[0]: [cd_data]}
    temp_data_dic   = {signals[0]: [temperatures]}

    return signal_data_dic, temp_data_dic, conditions, signals

def file_is_chirascan_thermal_ramp(cd_file):

    """
    Find if a file is of type Chirascan thermal ramp.

    Parameters
    ----------
    cd_file : str
        Path to the file

    Returns
    -------
    bool
        True if the file is a Chirascan thermal ramp, False otherwise

    """

    words_to_find = ["circulardichroism", "wavelength", "temperature"]

    with open(cd_file) as f:
        ls = f.read().splitlines()[:100]
        # Join the lines into a single string and convert to lowercase for easier searching
        text = " ".join(ls).lower()

        # Check if all the words are present in the text
        if all(word in text for word in words_to_find):
            return True

    return False

def read_chirascan_thermal_ramp(cd_file):

    """
    Read the CD data from a Chirascan file with a thermal ramp.

    Parameters
    ----------
    cd_file : str
        Path to the Chirascan file.

    Returns
    -------

    """

    # Open the file for reading
    with open(cd_file) as f:
        # Read all lines and split them into a list of lines
        ls = f.read().splitlines()

        # Find the row index where the numeric data starts (CD spectra)
        for i, l in enumerate(ls):
            # Check if the line contains "Wavelength"
            if "wavelength" in l.lower():
                # Assuming metadata starts two lines below the line containing "Wavelength"
                # Get the third row (i+2) and split it by comma to count the number of columns
                row2 = ls[i + 2].split(",")
                # Check if the first element of row2 is numeric, which indicates the start of data
                if row2[0].isnumeric():
                    # n_cols indicates the number of columns in the data
                    n_cols = len([x for x in row2 if x])
                    break

        # Initialize a list to store CD data, one list per row (wavelength)
        cd_data = []
        # Initialize an empty list to store wavelengths
        wavelength = []

        # Extract temperature values from the line after the "Wavelength" line
        temperature = ls[i + 1].split(",")[1:]
        temperature = [t for t in temperature if t]
        # Convert temperature values to a numpy array
        temperature = np.array(temperature).astype(float)

        # Iterate through the lines containing CD data (starting from row i+2)
        for i2, l in enumerate(ls[i + 2 :]):
            # Split the line using comma as the delimiter and remove any empty elements
            row = l.split(",")
            row = [x for x in row if x]

            # Check if the row has the correct number of columns (n_cols)
            if len(row) != n_cols:
                break

            # Append the wavelength value to the list
            wavelength.append(float(row[0]))

            # Convert
            signal_at_wavelength = np.array([parse_number(x) for x in row[1:]])

            cd_data.append(signal_at_wavelength)

        # Convert the list of wavelengths to a numpy array
        wavelength_data = np.array([str(wl) + " nm" for wl in wavelength])

        signal_data_dic ={}
        temp_data_dic  = {}

        for cd_signal,wl in zip(cd_data,wavelength_data):

            signal_data_dic[wl] = [cd_signal]
            temp_data_dic[wl] = [temperature]

        # extract file basename
        bname = os.path.basename(cd_file)

        conditions = [bname]

    return signal_data_dic, temp_data_dic, conditions, wavelength_data


"""
Tests to ensure that the main functionalities of the pychemelt Sample class work as expected.
The order of the tests is important, as some functions depend on the previous ones.
"""
import numpy as np
import pytest

from pychemelt import Monomer as Sample

sample = Sample()


def test_load_data():
    sample.read_multiple_files('./test_files/nDSFdemoFile.xlsx')

    assert len(sample.conditions) == 48
    assert len(sample.labels) == 48


def test_set_denaturant_concentrations():

    sample.set_denaturant_concentrations()

    assert np.min(sample.denaturant_concentrations_pre) == 0
    assert np.max(sample.denaturant_concentrations_pre) == 8.24

def test_select_signal():

    sample.set_signal(['350nm'])

    assert sample.signal_names == ['350nm']

def test_select_conditions():

    # Select without scaling
    sample.select_conditions(
        [False for _ in range(24)] + [True for _ in range(8)] + [False for _ in range(16)],
        normalise_to_global_max=False
    )

    assert len(sample.signal_lst_multiple) == 1
    assert len(sample.signal_lst_multiple[0]) == 8
    assert np.max(sample.signal_lst_multiple[0]) != 1.0

    # Select with scaling
    sample.select_conditions(
        [False for _ in range(24)] + [True for _ in range(8)] + [False for _ in range(16)],
        normalise_to_global_max=True
    )

    assert len(sample.signal_lst_multiple) == 1
    assert len(sample.signal_lst_multiple[0]) == 8
    assert np.max(sample.signal_lst_multiple[0]) == 100


def test_set_temperature_range():

    # Raise error if t_max < t_min
    with pytest.raises(ValueError):
        sample.set_temperature_range(80, 30)

    sample.set_temperature_range(30, 80)

    assert np.min(sample.temp_lst_multiple[0]) >= 30

    sample.set_temperature_range(5, 100)

    assert sample.user_min_temp == 5
    assert sample.user_max_temp == 100

def test_expand_signal():

    sample.expand_multiple_signal()

    assert sample.signal_lst_expanded is not None
    assert sample.temp_lst_expanded is not None

def test_set_signal_id():

    sample.set_signal_id()

    assert sample.signal_ids == [0 for _ in range(8)]

def test_estimate_derivative():

    sample.estimate_derivative()

    assert len(sample.deriv_lst_multiple[0]) == 8

def test_guess_Tm():

    sample.guess_Tm()

    # assert almost
    np.testing.assert_allclose(sample.t_melting_init_multiple[0][-1],63.82,rtol=0.01)

def test_guess_initial_parameters_ratio():

    sample.n_residues = 130
    sample.guess_initial_parameters(
        native_baseline_type='linear',
        unfolded_baseline_type='linear'
    )

    np.testing.assert_allclose(sample.thermodynamic_params_guess[0],71.8,rtol=0.1)

def test_different_format_data():

    file2 = './test_files/MX3005P.txt'

    sample.read_multiple_files([file2])

    assert len(sample.conditions) == 48

    assert  'Fluorescence' not in sample.signals
    assert '350nm' in sample.signals

def test_read_same_format_data():

    sample.read_multiple_files('./test_files/nDSFdemoFile.xlsx')

    assert len(sample.conditions) == 48 * 2





"""
Main class to handle thermal and chemical denaturation data
The current model assumes that the unfolding is reversible
"""

import pandas as pd
import numpy as np

from .utils.files  import (

    detect_file_type,
    load_nanoDSF_xlsx,
    load_panta_xlsx,
    load_uncle_multi_channel,
    load_thermofluor_xlsx,
    load_quantstudio_txt,
    load_mx3005p_txt,
    load_supr_dsf,
    load_csv_file,
    load_aunty_xlsx,
    read_jasco_thermal_ramp,
    read_jasco_thermal_ramp_format_2,
    read_chirascan_thermal_ramp

)

from .utils.math  import (
    temperature_to_celsius,
    is_evenly_spaced,
    first_derivative_savgol,
    constant_baseline,
    linear_baseline,
    quadratic_baseline,
    exponential_baseline
)

from .utils.processing import (
    subset_data,
    guess_Tm_from_derivative,
    clean_conditions_labels,
    subset_signal_by_temperature,
    estimate_signal_baseline_params, transform_to_list
)


class Sample:
    """
        Class to hold, process, and fit thermal and chemical denaturation data.

        This class manages multiple signal types (e.g., 350nm, 330nm, Ratio) and
        concentrations, providing an interface for global thermodynamic analysis
        under the assumption of reversible unfolding.



        Parameters
        ----------
        name : str, optional
            Identifier for the sample. Default is 'Test'.

        Attributes
        ----------
        signal_dic : dict
            Raw signal data mapped by signal name.
        temp_dic : dict
            Temperature data mapped by signal name.
        conditions : list of float
            Processed numeric values for experimental conditions (e.g., [Denaturant]).
        labels : list of str
            Original string labels for each condition.
        signals : list of str
            Names of all available signal types in the loaded files.
        nr_signals : int
            Number of distinct signal types selected for analysis.
        single_fit_done : bool
            Flag indicating if individual dataset fits have been completed.
        global_fit_done : bool
            Flag for global thermodynamic fitting with local baselines.
        global_global_fit_done : bool
            Flag for global thermodynamics and global baseline slopes.
        global_global_global_fit_done : bool
            Flag for global thermodynamics, slopes, and intercepts.
        """
    def __init__(self, name='Test'):

        self.name = name
        self.signal_dic = {}
        self.deriv_dic = {}
        self.temp_dic = {}
        self.conditions = []
        self.labels = []

        self.signals = []

        self.first_param_Ns_per_signal = []
        self.first_param_Us_per_signal = []
        self.second_param_Ns_per_signal = []
        self.second_param_Us_per_signal = []
        self.third_param_Ns_per_signal = []
        self.third_param_Us_per_signal = []

        self.Cp0 = 0

        self.signal_ids = None

        self.t_melting_init_multiple = None # Initial guess for tm based on derivative

        self.single_fit_done = False  # Individual thermodynamic parameters, baselines and slopes

        self.global_fit_done = False  # Global thermodynamic parameters, local baselines and slopes

        self.global_global_fit_done = False  # Global thermodynamic parameters, global slopes and local baselines

        self.global_global_global_fit_done = False  # Global thermodynamic parameters, global slopes and global baselines

        self.global_fit_params = None

        self.global_min_temp = 20
        self.global_max_temp = 100

        self.n_residues = 0

        self.pre_fit = True
        self.max_points = None

        self.user_min_temp = 5
        self.user_max_temp = 100

        self.predicted = None # Flattened list of fitted signals

        self.predicted_lst_multiple = None
        self.predicted_lst_multiple_scaled = None
        self.signal_lst_multiple_scaled = None

    def read_file(self, file):

        """
        Read the file and load the data into the sample object

        Parameters
        ----------
        file : str
            Path to the file

        Returns
        -------
        bool
            True if the file was read and loaded into the sample object
        """

        file_type = detect_file_type(file)

        read_fx_map = {
            'prometheus': load_nanoDSF_xlsx,
            'panta': load_panta_xlsx,
            'uncle': load_uncle_multi_channel,
            'thermofluor': load_thermofluor_xlsx,
            'quantstudio': load_quantstudio_txt,
            'mx3005p': load_mx3005p_txt,
            'supr': load_supr_dsf,
            'csv': load_csv_file,
            'aunty': load_aunty_xlsx,
            'jasco_thermal_ramp': read_jasco_thermal_ramp,
            'jasco_thermal_ramp_format_2': read_jasco_thermal_ramp_format_2,
            'chirascan_thermal_ramp':read_chirascan_thermal_ramp
        }

        read_fx = read_fx_map.get(file_type)

        signal_data_dic, temp_data_dic, conditions, signals_i = read_fx(file)

        # If self.signals is not empty, signals_i must be the same as self.signals
        if len(self.signals) > 0:
            if set(signals_i) != set(self.signals):
                # We can't combine files with different signals
                return False

        for si in signals_i:
            if si not in self.signals:
                self.signals.append(si)

        # For each key in signal_data_dic, find if the key already exists in self.signal_dic and append the data
        for k, v in signal_data_dic.items():
            # v is a list of arrays

            # make sure we have a list of arrays
            v = [np.array(x) for x in v]

            if k in self.signal_dic:
                self.signal_dic[k].extend(v)
            else:
                self.signal_dic[k] = v

        # For each key in temp_data_dic, find if the key already exists in self.temp_dic and append the data
        for k, v in temp_data_dic.items():

            min_temp_v = np.min(v)
            max_temp_v = np.max(v)

            self.global_min_temp = min(min_temp_v, self.global_min_temp)
            self.global_max_temp = max(max_temp_v, self.global_max_temp)

            # make sure we have a list of arrays
            v = [np.array(x) for x in v]

            if k in self.temp_dic:
                self.temp_dic[k].extend(v)
            else:
                self.temp_dic[k] = v

        # Keep original labels
        self.labels += conditions

        # Remove all non-numeric characters from the conditions
        conditions = clean_conditions_labels(conditions)

        self.conditions += conditions

        return True

    def read_multiple_files(self, files):
        """
        Read multiple files and load the data into the sample object

        Parameters
        ----------
        files : list or str
            List of paths to the files (or a single path)

        Returns
        -------
        bool
            True if the files were read and loaded into the sample object
        """

        # Convert to list if not isinstance(files, list):
        files = transform_to_list(files)

        for file in files:
            read_status = self.read_file(file)
            if not read_status:
                return False

        return True

    def set_signal(self, signal_names):

        """
        Set multiple signals to be used for the analysis.
        This way, we can fit globally multiple signals at the same time, such as 350nm and 330nm

        Parameters
        ----------
        signal_names : list or str
            List of names of the signals to be used. E.g., ['350nm','330nm'] or a single name

        Notes
        -----
        This method creates/updates the following attributes on the instance:
        - signal_lst_pre_multiple, temp_lst_pre_multiple : lists of lists
        - signal_names : list of signal name strings
        - nr_signals : int, number of signal types
        """

        # Convert signal_names to list if it is a string
        signal_names = transform_to_list(signal_names)

        signals = []
        temps = []

        for signal_name in signal_names:
            signal = self.signal_dic[signal_name]
            temp = self.temp_dic[signal_name]

            signals.append(signal)
            temps.append(temp)

        self.signal_lst_pre_multiple = signals
        self.temp_lst_pre_multiple = temps

        self.signal_names = signal_names
        self.nr_signals = len(signal_names)

        return None

    def set_temperature_range(
            self,
            min_temp=0,
            max_temp=100):
        """
        Set the temperature range for the sample

        Parameters
        ----------
        min_temp : float, optional
            Minimum temperature
        max_temp : float, optional
            Maximum temperature
        """

        # Give an error if max_temp is smaller than min_temp
        if max_temp < min_temp:
            raise ValueError('max_temp must be larger than min_temp')

        # Limit the signal to the temperature range
        for i in range(len(self.signal_lst_multiple)):
            self.signal_lst_multiple[i], self.temp_lst_multiple[i] = subset_signal_by_temperature(

                self.signal_lst_multiple[i],
                self.temp_lst_multiple[i],
                min_temp,
                max_temp

            )

        self.user_min_temp = min_temp
        self.user_max_temp = max_temp

        return None

    def set_signal_id(self):
        """
        Create a list with the same length as the total number of signals
        The elements of the list indicated the ID of the signal,
        e.g., all 350nm datasets are mapped to 0, all 330nm datasets to 1, etc.
        """

        signal_ids = []

        for i, s in enumerate(self.signal_lst_multiple):
            signal_ids.extend([i for _ in range(len(s))])

        self.signal_ids = signal_ids

        return None

    def estimate_derivative(self, window_length=8):

        """
        Estimate the derivative of the signal using Savitzky-Golay filter

        Parameters
        ----------
        window_length : int, optional
            Length of the filter window in degrees

        Notes
        -----
        Creates/updates attributes:
        - temp_deriv_lst_multiple, deriv_lst_multiple, deriv_lst_expanded : lists storing estimated derivatives and corresponding temps
        - predicted_deriv_lst_multiple : list storing estimated derivatives of predicted values
        """

        self.temp_deriv_lst_multiple = []
        self.deriv_lst_multiple = []
        self.deriv_lst_expanded = []

        for i in range(len(self.signal_lst_multiple)):

            temp_deriv_lst = []
            deriv_lst = []

            for s, t in zip(self.signal_lst_multiple[i], self.temp_lst_multiple[i]):

                check = is_evenly_spaced(t)

                if check:

                    derivative = first_derivative_savgol(t, s, window_length)

                else:

                    t_for_ip = np.arange(np.min(t), np.max(t), 0.1)

                    # We interpolate the data to make it evenly spaced, every 0.1 degrees
                    s = np.interp(t_for_ip, t, s)
                    t = t_for_ip

                    # We use the Savitzky-Golay filter to estimate the derivative
                    derivative = first_derivative_savgol(t, s, window_length)

                temp_deriv_lst.append(t)
                deriv_lst.append(derivative)
                self.deriv_lst_expanded.append(derivative)

            self.temp_deriv_lst_multiple.append(temp_deriv_lst)
            self.deriv_lst_multiple.append(deriv_lst)

        if self.max_points is not None:
            self.deriv_lst_expanded = [subset_data(x, self.max_points) for x in self.deriv_lst_expanded]

        if self.predicted_lst_multiple is not None:
            self.predicted_deriv_lst_multiple = []

            for i in range(len(self.predicted_lst_multiple)):

                temp_deriv_lst = []
                predicted_deriv_lst = []

                for p, t in zip(self.predicted_lst_multiple[i],
                                np.split(np.array(self.temp_lst_expanded), len(self.predicted_lst_multiple))[i]):

                    check = is_evenly_spaced(t)

                    if check:
                        derivative = first_derivative_savgol(t, p, window_length)

                    else:
                        t_for_ip = np.arange(np.min(t), np.max(t), 0.1)

                        # We interpolate the data to make it evenly spaced, every 0.1 degrees
                        p = np.interp(t_for_ip, t, p)
                        t = t_for_ip

                        # We use the Savitzky-Golay filter to estimate the derivative
                        derivative = first_derivative_savgol(t, p, window_length)

                    temp_deriv_lst.append(t)
                    predicted_deriv_lst.append(derivative)

                self.predicted_deriv_lst_multiple.append(predicted_deriv_lst)

        return None

    def guess_Tm(self, x1=6, x2=11):

        """
        Guess the Tm of the sample using the derivative of the signal

        Parameters
        ----------
        x1 : float, optional
            Shift from the minimum and maximum temperature to estimate the median of the initial and final baselines
        x2 : float, optional
            Shift from the minimum and maximum temperature to estimate the median of the initial and final baselines

        Notes
        -----
        x2 must be greater than x1.

        This method creates/updates attributes:
        - t_melting_init_multiple : list of initial Tm guesses per signal
        - t_melting_df_multiple : list of pandas.DataFrame objects with Tm vs Denaturant
        """

        self.t_melting_init_multiple = []
        self.t_melting_df_multiple = []

        for i in range(len(self.signal_lst_multiple)):
            t_melting_init = guess_Tm_from_derivative(
                self.temp_deriv_lst_multiple[i],
                self.deriv_lst_multiple[i],
                x1,
                x2
            )
            # Create a dataframe of the Tm values versus the denaturant concentrations
            t_melting_df = pd.DataFrame({
                'Tm': t_melting_init,
                'Denaturant': self.denaturant_concentrations
            })

            self.t_melting_df_multiple.append(t_melting_df)

            self.t_melting_init_multiple.append(t_melting_init)

        return None

    def estimate_baseline_parameters(
            self,
            native_baseline_type,
            unfolded_baseline_type,
            window_range_native=12,
            window_range_unfolded=12):

        """
        Estimate the baseline parameters for multiple signals

        Parameters
        ----------
        native_baseline_type : str
            one of 'constant', 'linear', 'quadratic', 'exponential'
        unfolded_baseline_type : str
            one of 'constant', 'linear', 'quadratic', 'exponential'
        window_range_native : int, optional
            Range of the window (in degrees) to estimate the baselines and slopes of the native state
        window_range_unfolded : int, optional
            Range of the window (in degrees) to estimate the baselines and slopes of the unfolded state

        Notes
        -----
        This method sets or updates these attributes:
        - bNs_per_signal, bUs_per_signal, kNs_per_signal, kUs_per_signal, qNs_per_signal, qUs_per_signal
        - poly_order_native, poly_order_unfolded
        """

        self.first_param_Ns_per_signal = []
        self.first_param_Us_per_signal = []
        self.second_param_Ns_per_signal = []
        self.second_param_Us_per_signal = []
        self.third_param_Ns_per_signal = []
        self.third_param_Us_per_signal = []

        for i in range(len(self.signal_lst_multiple)):
            p1Ns, p1Us, p2Ns, p2Us, p3Ns, p3Us = estimate_signal_baseline_params(
                self.signal_lst_multiple[i],
                self.temp_lst_multiple[i],
                native_baseline_type,
                unfolded_baseline_type,
                window_range_native,
                window_range_unfolded
            )

            self.first_param_Ns_per_signal.append(p1Ns)
            self.first_param_Us_per_signal.append(p1Us)
            self.second_param_Ns_per_signal.append(p2Ns)
            self.second_param_Us_per_signal.append(p2Us)
            self.third_param_Ns_per_signal.append(p3Ns)
            self.third_param_Us_per_signal.append(p3Us)

        baseline_fx_dic = {
            'constant': constant_baseline,
            'linear': linear_baseline,
            'quadratic': quadratic_baseline,
            'exponential': exponential_baseline
        }

        self.baseline_N_fx = baseline_fx_dic[native_baseline_type]
        self.baseline_U_fx = baseline_fx_dic[unfolded_baseline_type]

        self.native_baseline_type = native_baseline_type
        self.unfolded_baseline_type = unfolded_baseline_type

        return None

    def reset_fittings_results(self):
        """
        Deletes the results of previous fittings from the object

        """
        self.global_fit_done = False  # Global thermodynamic parameters, local baselines and slopes
        self.global_global_fit_done = False  # Global thermodynamic parameters, global slopes and local baselines
        self.global_global_global_fit_done = False  # Global thermodynamic parameters, global slopes and global baselines

        self.first_param_Ns_per_signal = []
        self.first_param_Us_per_signal = []
        self.second_param_Ns_per_signal = []
        self.second_param_Us_per_signal = []
        self.third_param_Ns_per_signal = []
        self.third_param_Us_per_signal = []

        self.params_df = None
        self.dg_df = None

        self.global_fit_params = None
        self.predicted_lst_multiple = None

        self.predicted_lst_multiple_scaled = None
        self.signal_lst_multiple_scaled = None

        self.p0 = None
        self.low_bounds = None
        self.high_bounds = None
        self.rel_errors = None

        return None

    def expand_multiple_signal(self):

        """
        Create a single list with all the signals
        Create a single list with all the temperatures

        Notes
        -----
        Creates/updates attributes:
        - signal_lst_expanded, temp_lst_expanded
        - signal_lst_expanded_subset, temp_lst_expanded_subset
        """

        # Create a single list with all the signals
        self.signal_lst_expanded = []
        self.temp_lst_expanded = []

        for i in range(len(self.signal_lst_multiple)):
            self.signal_lst_expanded += self.signal_lst_multiple[i]
            self.temp_lst_expanded += self.temp_lst_multiple[i]

        # Create a reduced dataset for faster fitting
        self.signal_lst_expanded_subset = [subset_data(x, 60) for x in self.signal_lst_expanded]
        self.temp_lst_expanded_subset = [subset_data(x, 60) for x in self.temp_lst_expanded]


        if self.max_points is not None:
            self.signal_lst_expanded = [subset_data(x, self.max_points) for x in self.signal_lst_expanded]
            self.temp_lst_expanded = [subset_data(x, self.max_points) for x in self.temp_lst_expanded]


        return None

    def create_params_df(self):
        """
        Create a dataframe of the parameters
        """

        # convert the first param to Celsius
        self.global_fit_params[0] = temperature_to_celsius(self.global_fit_params[0])
        self.low_bounds[0] = temperature_to_celsius(self.low_bounds[0])
        self.high_bounds[0] = temperature_to_celsius(self.high_bounds[0])


        # Create a dataframe of the parameters
        self.params_df = pd.DataFrame({
            'Parameter': self.params_names,
            'Value': self.global_fit_params,
            'Relative error (%)': self.rel_errors,
            'Fitting low Bound': self.low_bounds,
            'Fitting high Bound': self.high_bounds
        })

        return None


"""
Module containing functions to perform Singular Value Decomposition (SVD) and Principal Component Analysis (PCA)
on spectral data, along with utilities for manipulating basis spectra and coefficients.

Author: Osvaldo Burastero
"""

import numpy as np


def apply_svd(X):
    """
    Perform Singular Value Decomposition (SVD) on the input data matrix X.

    Parameters
    ----------
        X : numpy array of shape (n_wavelengths, n_measurements)
            The input data matrix to decompose.

    Returns
    -------
        explained_variance : numpy array
            The cumulative explained variance for each component.
        basis_spectra     : numpy array
            The left singular vectors (U matrix) representing the basis spectra.
        coefficients      : numpy array
            The coefficients associated with each basis spectrum.
    """

    U, S, _ = np.linalg.svd(X)

    # Calculate the total variance or correlation
    total_variance = np.sum(S ** 2)
    cumulative_variance = np.cumsum(S ** 2)

    # The matrix V contains the variation of each component against the temperature / measurement dimension

    a_is = []

    for i in range(U.shape[1]):
        def coefficients_bi(column):
            return U[:, i].dot(column)

        a_i = np.apply_along_axis(coefficients_bi, axis=0, arr=X)

        a_is.append(a_i)

    coefficients = np.array(a_is)

    # Basis spectra
    basis_spectra = U

    # Cumulated explained variance of the components
    explained_variance = cumulative_variance / total_variance * 100

    return explained_variance, basis_spectra, coefficients


def filter_basis_spectra(explained_variance, basis_spectra_all, coefficients_all, explained_variance_threshold=99):
    """
    Filter the basis spectra and coefficients based on the explained variance threshold
    Parameters
    ----------
        explained_variance         : numpy array
                                     The cumulative explained variance for each component.
        basis_spectra_all         : numpy array
                                     The left singular vectors (U matrix) representing the basis spectra.
        coefficients_all          : numpy array
                                     The coefficients associated with each basis spectrum.
        explained_variance_threshold : float, optional
                                     The threshold for explained variance to filter components. Default is 99.
    Returns
    -------
        basis_spectra  : numpy array
                         The filtered basis spectra.
        coefficients   : numpy array
                         The filtered coefficients.
        k              : int
                         The number of components that meet the explained variance threshold.
    """

    # Find the number of components (k) that capture at least threshold*100 percent of the variance or correlation
    k = np.sum(explained_variance < explained_variance_threshold) + 1

    basis_spectra = basis_spectra_all[:, :k]
    coefficients = coefficients_all[:k, :]

    return basis_spectra, coefficients, k


def align_basis_spectra_and_coefficients(X, basis_spectra, coefficients):
    """
    Align the basis spectra peaks to the original data
    Parameters
    ----------
        X              : numpy array of shape (n_samples, n_features)
                         The input data matrix.
        basis_spectra  : numpy array
                         The basis spectra obtained from SVD.
        coefficients   : numpy array
                         The coefficients associated with each basis spectrum.
    Returns
    -------
        basis_spectra  : numpy array
                         The aligned basis spectra.
        coefficients   : numpy array
                         The adjusted coefficients.

    """

    # Align basis spectra peaks to the original data
    # In other words, we want that if the original spectra has a peak with positive values of the CD signal,
    # so does our basis spectra

    # Fix the n_cutoff to remove the first n and last n rows of X before finding the peak

    n_cutoff = 5

    maxV_abs = np.abs(np.max(X[n_cutoff:-n_cutoff, :]))
    minV_abs = np.abs(np.min(X[n_cutoff:-n_cutoff, :]))

    positive_peak = maxV_abs > minV_abs

    k = basis_spectra.shape[1]

    for i in range(k):

        prcomp_i = basis_spectra[:, i]

        maxV_abs_prcomp_i = np.abs(np.max(prcomp_i[n_cutoff:-n_cutoff]))
        minV_abs_prcomp_i = np.abs(np.min(prcomp_i[n_cutoff:-n_cutoff]))

        positive_peak_prcomp_i = maxV_abs_prcomp_i > minV_abs_prcomp_i

        if positive_peak_prcomp_i != positive_peak:
            coeff_i = coefficients[i, :]

            basis_spectra[:, i] = - prcomp_i
            coefficients[i, :] = - coeff_i

    return basis_spectra, coefficients


def angle_from_cathets(adjacent_leg, opposite_leg):
    """
    Calculate the angle between the hypotenuse and the adjacent leg of a right triangle.
    Parameters
    ----------
        adjacent_leg : float
            Length of the adjacent leg.
        opposite_leg : float
            Length of the opposite leg.
    Returns
    -------
        angle_in_radians : float
            Angle in radians between the hypotenuse and the adjacent leg.
    """

    hypotenuse = np.sqrt(adjacent_leg ** 2 + opposite_leg ** 2)

    return np.arccos(adjacent_leg / hypotenuse)


def get_2d_counterclockwise_rot_matrix(angle_in_radians):
    """
    Obtain the rotation matrix for a 2d coordinates system using a counterclockwise direction
    Parameters
    ----------
        angle_in_radians : float
            Angle in radians for the rotation.
    Returns
    -------
        rotM : numpy array
            2x2 rotation matrix.
    """

    rotM = np.array([[np.cos(angle_in_radians), np.sin(angle_in_radians)],
                     [-np.sin(angle_in_radians), np.cos(angle_in_radians)]])

    return rotM


def get_3d_counterclockwise_rot_matrix_around_z_axis(angle_in_radians):
    """
    Obtain the rotation matrix for a 3d coordinates system around the z axis using a counterclockwise direction
    Parameters
    ----------
        angle_in_radians : float
            Angle in radians for the rotation.
    Returns
    -------
        rotM : numpy array
            3x3 rotation matrix.
    """

    rotM = np.array([[np.cos(angle_in_radians), np.sin(angle_in_radians), 0],
                     [-np.sin(angle_in_radians), np.cos(angle_in_radians), 0],
                     [0, 0, 1]])

    return rotM


def get_3d_clockwise_rot_matrix_around_y_axis(angle_in_radians):
    """
    Obtain the rotation matrix for a 3d coordinates system around the y axis using a clockwise direction
    Parameters
    ----------
        angle_in_radians : float
            Angle in radians for the rotation.
    Returns
    -------
        rotM : numpy array
            3x3 rotation matrix.
    """

    rotM = np.array([[np.cos(angle_in_radians), 0, np.sin(angle_in_radians)],
                     [0, 1, 0],
                     [-np.sin(angle_in_radians), 0, np.cos(angle_in_radians)]])

    return rotM


def rotate_two_basis_spectra(X, basis_spectra, pca_based=False):
    """
    Create a new basis spectra using a linear combination of the first and second basis spectra

    Parameters
    ----------
        X : numpy array
            The raw data matrix of size n*m, where 'n' is the number of measured wavelengths
            and 'm' is the number of acquired spectra.
        basis_spectra : numpy array
            The matrix containing the set of basis spectra.
        pca_based : bool, optional
            Boolean to decide if we need to center the matrix X. Default is False.

    Returns
    -------
        basis_spectra_new : numpy array
            The new set of basis spectra.
        coefficients : numpy array
            The new set of associated coefficients.
    """

    if pca_based:
        X_mean = np.mean(X, axis=1, keepdims=True)
        X = X - X_mean

    first_spectrum = X[:, 0]

    c1 = first_spectrum.dot(basis_spectra[:, 0])
    c2 = first_spectrum.dot(basis_spectra[:, 1])

    rotAngle = angle_from_cathets(c1, c2)

    rotM = get_2d_counterclockwise_rot_matrix(rotAngle)

    basis_spectra_new = np.dot(basis_spectra[:, :2], rotM)
    coefficients = np.dot(basis_spectra_new.T, X)

    return basis_spectra_new, coefficients


def rotate_three_basis_spectra(X, basis_spectra, pca_based=False):
    """
    Create a new basis spectra using a linear combination from the first, second and third basis spectra

    Parameters
    ----------
        X : numpy array
            The raw data matrix of size n*m, where 'n' is the number of measured wavelengths
            and 'm' is the number of acquired spectra.
        basis_spectra : numpy array
            The matrix containing the set of basis spectra.
        pca_based : bool, optional
            Boolean to decide if we need to center the matrix X. Default is False.

    Returns
    -------
        basis_spectra_new : numpy array
            The new set of basis spectra.
        coefficients_subset : numpy array
            The new set of associated coefficients.

    """

    if pca_based:
        X_mean = np.mean(X, axis=1, keepdims=True)
        X = X - X_mean

    first_spectrum = X[:, 0]

    c1 = first_spectrum.dot(basis_spectra[:, 0])
    c2 = first_spectrum.dot(basis_spectra[:, 1])
    c3 = first_spectrum.dot(basis_spectra[:, 2])

    zAngle = angle_from_cathets(c1, c2)
    yAngle = angle_from_cathets(np.sqrt(c1 ** 2 + c2 ** 2), c3)

    rotZaxis = get_3d_counterclockwise_rot_matrix_around_z_axis(zAngle)
    rotYaxis = get_3d_clockwise_rot_matrix_around_y_axis(yAngle)

    basis_z_rot = np.dot(basis_spectra[:, :3], rotZaxis)
    basis_spectra_new = np.dot(basis_z_rot, rotYaxis)
    coefficients = np.dot(basis_spectra_new.T, X)

    return basis_spectra_new, coefficients


def reconstruct_spectra(basis_spectra, coefficients, X=None, pca_based=False):
    """
    Reconstruct the original spectra based on the set of basis spectra and the associated coefficients

    Parameters
    ----------
        basis_spectra       : numpy array
            The matrix containing the set of basis spectra.
        coefficients        : numpy array
            The associated coefficients of each basis spectrum.
        X                   : numpy array, optional
            Only used if pca_based equals TRUE!
            X is the raw data matrix of size n*m, where
            'n' is the number of measured wavelengths and
            'm' is the number of acquired spectra.
        pca_based           : bool, optional
            Boolean to decide if we need to extract the mean from the the X raw data matrix. Default is False.

        Returns
        -------
        fitted : numpy array
            The reconstructed matrix which should be close the original raw data.
    """

    fitted = (basis_spectra @ coefficients)

    # Add the mean, if needed
    if pca_based:
        X_mean = np.mean(X, axis=1, keepdims=True)
        fitted = fitted + X_mean

    return fitted


def explained_variance_from_orthogonal_vectors(vectors, coefficients, total_variance):
    """
    Useful to get the percentage of variance, not in the coordinate space provided by PCA/SVD,
    but against a different set of (rotated) vectors.

    Parameters
    ----------
        vectors        : numpy array
            The set of orthogonal vectors.
        coefficients   : numpy array
            The associated coefficients of each orthogonal vector.
        total_variance : float
            The total variance of the original data (mean subtracted if we performed PCA...).

    Returns
    -------
        explained_variance : list
            The amount of explained variance by each orthogonal vector.
    """

    explained_variance = []

    for i in range(vectors.shape[1]):
        a = np.linalg.norm(coefficients[i, :]) ** 2
        b = np.linalg.norm(vectors[:, i]) ** 2

        explained_variance.append(a / b)

    return 100 * np.cumsum(explained_variance) / total_variance


def apply_pca(X):
    """
    Perform Principal Component Analysis (PCA) on the input data matrix X.
    Parameters
    ----------
        X : numpy array of shape (n_wavelengths, n_measurements)
            The input data matrix to decompose.
    Returns
    -------
        cum_sum_eigenvalues : numpy array
            The cumulative explained variance for each principal component.
        principal_components : numpy array
            The principal components (eigenvectors) representing the basis spectra.
        coefficients : numpy array
            The coefficients associated with each principal component.
    """

    X = X.T  # We need to transpose X to have samples as rows and features as columns

    X_mean = np.mean(X, axis=0)
    X = X - X_mean

    # compute the covariance matrix
    cov_mat = np.cov(X, rowvar=False)

    # find the eigen vectors and associated eigen values
    eigen_values, eigen_vectors = np.linalg.eigh(cov_mat)

    # sort the eigenvalues in descending order
    sorted_index = np.argsort(eigen_values)[::-1]

    sorted_eigenvalue = eigen_values[sorted_index]

    # similarly sort the eigenvectors
    sorted_eigenvectors = eigen_vectors[:, sorted_index]

    # compute the total variance
    total_eigenvalues = np.sum(sorted_eigenvalue)

    # compute the explained variance
    exp_var_pca = (sorted_eigenvalue / total_eigenvalues * 100)

    # compute the cumulative explained variance
    cum_sum_eigenvalues = np.cumsum(exp_var_pca)

    principal_components = sorted_eigenvectors

    a_is = []

    for i in range(principal_components.shape[1]):
        def coefficients_bi(column):
            # Your custom logic here
            return principal_components[:, i].dot(column)

        a_i = np.apply_along_axis(coefficients_bi, axis=1, arr=X)

        a_is.append(a_i)

    coefficients = np.array(a_is)

    return cum_sum_eigenvalues, principal_components, coefficients


def recalc_explained_variance(basis_spectra, coefficients, X, pca_based=False):
    """
    Recalculate the explained variance of a set of basis spectra and associated coefficients
    Parameters
    ----------
        basis_spectra : numpy array
                         The basis spectra.
        coefficients  : numpy array
                         The associated coefficients of each basis spectrum.
        X             : numpy array
                         The raw data matrix of size n*m, where 'n' is the number of measured wavelengths
                         and 'm' is the number of acquired spectra.
        pca_based     : bool, optional
                         Boolean to decide if we need to center the matrix X. Default is False.
    Returns
    -------
        explained_variance : numpy array
                             The cumulative explained variance for each component.
    """

    if pca_based:
        X_mean = np.mean(X, axis=1, keepdims=True)
        X = X - X_mean

    total_variance = np.linalg.norm(X) ** 2  # Total variance in the data

    explained_variance = explained_variance_from_orthogonal_vectors(
        basis_spectra,
        coefficients,
        total_variance)

    return explained_variance


import numpy as np

from pychemelt.utils.fitting import evaluate_need_to_refit

# Model / ground-truth parameters
DHm_VAL = 120
Tm_VAL = 65
CP0_VAL = 1.8
M0_VAL = 2.6
M1_VAL = 0

INTERCEPT_N = 100
PRE_EXP_N = 1
C_N_VAL = 1
ALPHA_N_VAL = 0.1
INTERCEPT_U = 110
PRE_EXP_U = 1
C_U_VAL = 1
ALPHA_U_VAL = 0.2

PARAMS = [Tm_VAL,DHm_VAL,CP0_VAL,M0_VAL,M1_VAL,INTERCEPT_N,INTERCEPT_U,C_N_VAL,C_U_VAL,ALPHA_N_VAL,ALPHA_U_VAL,PRE_EXP_N,PRE_EXP_U]
LOW_BOUNDS = [-np.inf for _ in PARAMS]
UPP_BOUNDS = [np.inf for _ in PARAMS]

def test_evaluate_need_to_refit():

    re_fit, p0, low_bounds, high_bounds = evaluate_need_to_refit(
        PARAMS,
        UPP_BOUNDS,
        LOW_BOUNDS,
        PARAMS,
        fit_m1=True,
        check_cp=True,
        check_dh=True,
        check_tm=True,
        fixed_cp=False
    )

    assert p0 == PARAMS and not re_fit

def test_evaluate_need_to_refit_tm():

    upp_bounds_tm = UPP_BOUNDS.copy()
    upp_bounds_tm[0] = 70

    re_fit, p0, low_bounds, high_bounds = evaluate_need_to_refit(
        PARAMS,
        upp_bounds_tm,
        LOW_BOUNDS,
        PARAMS,
        fit_m1=True,
        check_cp=False,
        check_dh=False,
        check_tm=True,
        fixed_cp=False
    )

    assert high_bounds[0] == (Tm_VAL + 12) and re_fit

    low_bounds_tm = LOW_BOUNDS.copy()
    low_bounds_tm[0] = 60

    re_fit, p0, low_bounds, high_bounds = evaluate_need_to_refit(
        PARAMS,
        UPP_BOUNDS,
        low_bounds_tm,
        PARAMS,
        fit_m1=True,
        check_cp=False,
        check_dh=False,
        check_tm=True,
        fixed_cp=False
    )

    assert low_bounds[0] == (Tm_VAL - 12) and re_fit

def test_evaluate_need_to_refit_dh():

    upp_bounds_dh = UPP_BOUNDS.copy()
    upp_bounds_dh[1] = 130

    re_fit, p0, low_bounds, high_bounds = evaluate_need_to_refit(
        PARAMS,
        upp_bounds_dh,
        LOW_BOUNDS,
        PARAMS,
        fit_m1=True,
        check_cp=False,
        check_dh=True,
        check_tm=False,
        fixed_cp=False
    )

    assert high_bounds[1] == (DHm_VAL + 80) and re_fit

def test_evaluate_need_to_refit_cp():

    upp_bounds_dh = UPP_BOUNDS.copy()
    upp_bounds_dh[2] = 1.9

    re_fit, p0, low_bounds, high_bounds = evaluate_need_to_refit(
        PARAMS,
        upp_bounds_dh,
        LOW_BOUNDS,
        PARAMS,
        fit_m1=True,
        check_cp=True,
        check_dh=False,
        check_tm=False,
        fixed_cp=False
    )

    assert high_bounds[2] == (CP0_VAL + 1) and re_fit

def test_evaluate_need_to_fixed_cp():

    re_fit, p0, low_bounds, high_bounds = evaluate_need_to_refit(
        PARAMS,
        UPP_BOUNDS,
        LOW_BOUNDS,
        PARAMS,
        fit_m1=True,
        check_cp=True,
        check_dh=False,
        check_tm=False,
        fixed_cp=True
    )

    assert not re_fit

def test_evaluate_need_to_refit_m0():

    upp_bounds_m0 = UPP_BOUNDS.copy()
    upp_bounds_m0[3] = 3

    re_fit, p0, low_bounds, high_bounds = evaluate_need_to_refit(
        PARAMS,
        upp_bounds_m0,
        LOW_BOUNDS,
        PARAMS,
        fit_m1=True,
        check_cp=False,
        check_dh=False,
        check_tm=False,
        fixed_cp=False
    )

    assert high_bounds[3] == (M0_VAL + 2) and re_fit


def test_evaluate_need_to_refit_m1():

    upp_bounds_m1 = UPP_BOUNDS.copy()
    upp_bounds_m1[4] = 0.01

    re_fit, p0, low_bounds, high_bounds = evaluate_need_to_refit(
        PARAMS,
        upp_bounds_m1,
        LOW_BOUNDS,
        PARAMS,
        fit_m1=True,
        check_cp=False,
        check_dh=False,
        check_tm=False,
        fixed_cp=False
    )

    assert high_bounds[4] == (M1_VAL + 1) and re_fit

    low_bounds_m1 = LOW_BOUNDS.copy()
    low_bounds_m1[4] = -0.01

    re_fit, p0, low_bounds, high_bounds = evaluate_need_to_refit(
        PARAMS,
        UPP_BOUNDS,
        low_bounds_m1,
        PARAMS,
        fit_m1=True,
        check_cp=False,
        check_dh=False,
        check_tm=False,
        fixed_cp=False
    )

    assert low_bounds[4] == (M1_VAL - 1) and re_fit

def test_evaluate_need_to_refit_other_params():

    upp_bounds_in = UPP_BOUNDS.copy()
    upp_bounds_in[6:] = [x + 0.001 for x in PARAMS[6:]]

    low_bounds_in = LOW_BOUNDS.copy()
    low_bounds_in[6:] = [x - 0.001 for x in PARAMS[6:]]

    re_fit, p0, low_bounds_out, upp_bounds_out = evaluate_need_to_refit(
        PARAMS,
        upp_bounds_in,
        low_bounds_in,
        PARAMS,
        fit_m1=True,
        check_cp=False,
        check_dh=False,
        check_tm=False,
        fixed_cp=False
    )

    assert re_fit

    for i in range(6,len(PARAMS)):

        assert low_bounds_out[i] == low_bounds_in[i] / 50
        assert upp_bounds_out[i] == upp_bounds_in[i] * 50

